/*
    File:		AJZaurusUSB.cpp
	
    Description:	This is an experimental driver that enables Mac OS X to communicate with a Sharp Zaurus
	                    over IP over USB. It is based on the USBCDCEthernet example which is
						Copyright 1998-2002 Apple Computer, Inc. All rights reserved.
						Portions have been ported from usbdnet.c (Linux driver) written by
						Stuart Lynne <sl@lineo.com> and Tom Rushworth <tbr@lineo.com> with some
						algorithms adopted from usbnet.c written by
						David Brownell <dbrownell@users.sourceforge.net>.
					
					Original description for USBCDCEthernet:
  	                    This is a sample USB Communication Device Class (CDC) driver, Ethernet model.
                        Note that this sample has not been tested against any actual hardware since there
                        are very few CDC Ethernet devices currently in existence. 
                        
                        This sample requires Mac OS X 10.1 and later. If built on a version prior to 
                        Mac OS X 10.2, a compiler warning "warning: ANSI C++ forbids data member `ip_opts'
                        with same name as enclosing class" will be issued. This warning can be ignored.
                        
    Copyright:		Copyright 2002, 2003 Andreas Junghans
	
    Disclaimer:		This program is free software; you can redistribute it and/or modify
					it under the terms of the GNU General Public License as published by
					the Free Software Foundation; either version 2, or (at your option)
					any later version.

					This program is distributed in the hope that it will be useful,
					but WITHOUT ANY WARRANTY; without even the implied warranty of
					MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
					GNU General Public License for more details.

					You should have received a copy of the GNU General Public License
					along with this program; if not, write to the Free Software
					Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
				
	Change History (most recent first):
        
			<7>  	07/04/03	Support for Zaurus C760.
			<6>  	02/18/03	Increased stability + support for iPaq/Opie.
			<5>  	02/10/03	Bug fix (introduces a workloop).
			<4>  	01/24/03	Support for more Zaurus models.
			<3>  	01/11/03	Modified to include latest changes from sample driver
								(provided by Russ Winsper <russw@apple.com>)
			<2>  	12/22/02	Modified to support the Sharp Zaurus (and some bug fixes)
            <1>	 	11/08/02	New sample (provided by Apple Computer, Inc.)
        
*/

#include "AJZaurusUSB.h"

#define MIN_BAUD (50 << 1)

static globals	g;	// Instantiate the globals
    
static struct MediumTable
{
    UInt32	type;
    UInt32	speed;
}

mediumTable[] =
{
    {kIOMediumEthernetNone,												0},
    {kIOMediumEthernetAuto,												0},
    {kIOMediumEthernet10BaseT 	 | kIOMediumOptionHalfDuplex,								10},
    {kIOMediumEthernet10BaseT 	 | kIOMediumOptionFullDuplex,								10},
    {kIOMediumEthernet100BaseTX  | kIOMediumOptionHalfDuplex,								100},
    {kIOMediumEthernet100BaseTX  | kIOMediumOptionFullDuplex,								100}
};

#define	numStats	13
UInt16	stats[13] = {	kXMIT_OK_REQ,
                        kRCV_OK_REQ,
                        kXMIT_ERROR_REQ,
                        kRCV_ERROR_REQ, 
                        kRCV_CRC_ERROR_REQ,
                        kRCV_ERROR_ALIGNMENT_REQ,
                        kXMIT_ONE_COLLISION_REQ,
                        kXMIT_MORE_COLLISIONS_REQ,
                        kXMIT_DEFERRED_REQ,
                        kXMIT_MAX_COLLISION_REQ,
                        kRCV_OVERRUN_REQ,
                        kXMIT_TIMES_CARRIER_LOST_REQ,
                        kXMIT_LATE_COLLISIONS_REQ
                    };

#define super IOEthernetController

OSDefineMetaClassAndStructors(org_lucidcake_driver_AJZaurusUSB, IOEthernetController);

#if USE_ELG
/****************************************************************************************************/
//
//		Function:	AllocateEventLog
//
//		Inputs:		size - amount of memory to allocate
//
//		Outputs:	None
//
//		Desc:		Allocates the event log buffer
//
/****************************************************************************************************/

static void AllocateEventLog(UInt32 size)
{
    if (g.evLogBuf)
        return;

    g.evLogFlag = 0;            // assume insufficient memory
    g.evLogBuf = (UInt8*)IOMalloc(size);
    if (!g.evLogBuf)
    {
        kprintf("org_lucidcake_driver_AJZaurusUSB evLog allocation failed ");
        return;
    }

    bzero(g.evLogBuf, size);
    g.evLogBufp	= g.evLogBuf;
    g.evLogBufe	= g.evLogBufp + kEvLogSize - 0x20; // ??? overran buffer?
    g.evLogFlag  = 0xFEEDBEEF;	// continuous wraparound
//	g.evLogFlag  = 'step';		// stop at each ELG
//	g.evLogFlag  = 0x0333;		// any nonzero - don't wrap - stop logging at buffer end

    IOLog("AllocateEventLog - &globals=%8x buffer=%8x", (unsigned int)&g, (unsigned int)g.evLogBuf);

    return;
	
}/* end AllocateEventLog */

/****************************************************************************************************/
//
//		Function:	EvLog
//
//		Inputs:		a - anything, b - anything, ascii - 4 charater tag, str - any info string			
//
//		Outputs:	None
//
//		Desc:		Writes the various inputs to the event log buffer
//
/****************************************************************************************************/

static void EvLog(UInt32 a, UInt32 b, UInt32 ascii, char* str)
{
    register UInt32	*lp;           // Long pointer
    mach_timespec_t	time;

    if (g.evLogFlag == 0)
        return;

    IOGetTime(&time);

    lp = (UInt32*)g.evLogBufp;
    g.evLogBufp += 0x10;

    if (g.evLogBufp >= g.evLogBufe)       // handle buffer wrap around if any
    {    
        g.evLogBufp  = g.evLogBuf;
        if (g.evLogFlag != 0xFEEDBEEF)    // make 0xFEEDBEEF a symbolic ???
            g.evLogFlag = 0;                // stop tracing if wrap undesired
    }

        // compose interrupt level with 3 byte time stamp:

    *lp++ = (g.intLevel << 24) | ((time.tv_nsec >> 10) & 0x003FFFFF);   // ~ 1 microsec resolution
    *lp++ = a;
    *lp++ = b;
    *lp   = ascii;

    if(g.evLogFlag == 'step')
    {	
        static char	code[ 5 ] = {0,0,0,0,0};
        *(UInt32*)&code = ascii;
        IOLog("%8x org_lucidcake_driver_AJZaurusUSB: %8x %8x %s\n", time.tv_nsec>>10, (unsigned int)a, (unsigned int)b, code);
    }

    return;
	
}/* end EvLog */
#endif // USE_ELG

#if LOG_DATA
/****************************************************************************************************/
//
//		Function:	Asciify
//
//		Inputs:		i - the nibble
//
//		Outputs:	return byte - ascii byte
//
//		Desc:		Converts to ascii. 
//
/****************************************************************************************************/
 
static UInt8 Asciify(UInt8 i)
{

    i &= 0xF;
    if (i < 10)
        return('0' + i);
    else return(55  + i);
	
}/* end Asciify */

#define dumplen		32		// Set this to the number of bytes to dump and the rest should work out correct

#define buflen		((dumplen*2)+dumplen)+3
#define Asciistart	(dumplen*2)+3

/****************************************************************************************************/
//
//		Function:	USBLogData
//
//		Inputs:		Dir - direction
//				Count - number of bytes
//				buf - the data
//
//		Outputs:	
//
//		Desc:		Puts the data in the log. 
//
/****************************************************************************************************/

static void USBLogData(UInt8 Dir, UInt32 Count, char *buf)
{
    UInt8	wlen, i, Aspnt, Hxpnt;
    UInt8	wchr;
    char	LocBuf[buflen+1];

    for (i=0; i<=buflen; i++)
    {
        LocBuf[i] = 0x20;
    }
    LocBuf[i] = 0x00;
	
    if (Dir == kUSBIn)
    {
        IOLog("org_lucidcake_driver_AJZaurusUSB: USBLogData - Read Complete, size = %8x\n", Count);
    } else {
        if (Dir == kUSBOut)
        {
            IOLog("org_lucidcake_driver_AJZaurusUSB: USBLogData - Write, size = %8x\n", Count);
        } else {
            if (Dir == kUSBAnyDirn)
            {
                IOLog("org_lucidcake_driver_AJZaurusUSB: USBLogData - Other, size = %8x\n", Count);
            }
        }			
    }

    if (Count > dumplen)
    {
        wlen = dumplen;
    } else {
        wlen = Count;
    }
	
    if (wlen > 0)
    {
        Aspnt = Asciistart;
        Hxpnt = 0;
        for (i=1; i<=wlen; i++)
        {
            wchr = buf[i-1];
            LocBuf[Hxpnt++] = Asciify(wchr >> 4);
            LocBuf[Hxpnt++] = Asciify(wchr);
            if ((wchr < 0x20) || (wchr > 0x7F)) 		// Non printable characters
            {
                LocBuf[Aspnt++] = 0x2E;				// Replace with a period
            } else {
                LocBuf[Aspnt++] = wchr;
            }
        }
        LocBuf[(wlen + Asciistart) + 1] = 0x00;
        IOLog(LocBuf);
        IOLog("\n");
        IOSleep(Sleep_Time);					// Try and keep the log from overflowing
    } else {
        IOLog("org_lucidcake_driver_AJZaurusUSB: USBLogData - No data, Count=0\n");
    }
	
}/* end USBLogData */
#endif // LOG_DATA


// AJ: begin CRC code (ported from Linux driver usbdnet.c written by
// Stuart Lynne <sl@lineo.com> and Tom Rushworth <tbr@lineo.com>
// with some algorithms adopted from usbnet.c by
// David Brownell <dbrownell@users.sourceforge.net>
// -->
UInt32 crc32_table[256] = {
    0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
    0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
    0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
    0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
    0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
    0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
    0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
    0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
    0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
    0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
    0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
    0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
    0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
    0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
    0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
    0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
    0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
    0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
    0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
    0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
    0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
    0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
    0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
    0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
    0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
    0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
    0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
    0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
    0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
    0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
    0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
    0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

#define CRC32_INITFCS     0xffffffff  // Initial FCS value 
#define CRC32_GOODFCS     0xdebb20e3  // Good final FCS value 

#define CRC32_FCS(fcs, c) (((fcs) >> 8) ^ crc32_table[((fcs) ^ (c)) & 0xff])

/* fcs_memcpy32 - memcpy and calculate fcs
 * Perform a memcpy and calculate fcs using ppp 32bit CRC algorithm.
 */ 
static inline UInt32 fcs_memcpy32(unsigned char *dp, unsigned char *sp, int len, UInt32 fcs)
{   
    for (;len-- > 0; fcs = CRC32_FCS(fcs, *dp++ = *sp++));
    return fcs;
}

/* fcs_pad32 - pad and calculate fcs
 * Pad and calculate fcs using ppp 32bit CRC algorithm.
 */
static inline UInt32 fcs_pad32(unsigned char *dp, int len, UInt32 fcs)
{
    for (;len-- > 0; fcs = CRC32_FCS(fcs, *dp++ = '\0'));
    return fcs;
}

/* fcs_compute32 - memcpy and calculate fcs
 * Perform a memcpy and calculate fcs using ppp 32bit CRC algorithm.
 */
static inline UInt32 fcs_compute32(unsigned char *sp, int len, UInt32 fcs)
{
    for (;len-- > 0; fcs = CRC32_FCS(fcs, *sp++));
    return fcs;
}
// <--


/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::commReadComplete
//
//		Inputs:		obj - me, param - parameter block(the Port), rc - return code, remaining - what's left
//												(whose idea was that?)
//
//		Outputs:	None
//
//		Desc:		Interrupt pipe (Comm interface) read completion routine
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::commReadComplete(void *obj, void *param, IOReturn rc, UInt32 remaining)
{
    org_lucidcake_driver_AJZaurusUSB	*me = (org_lucidcake_driver_AJZaurusUSB*)obj;
    IOReturn		ior;
    UInt32		dLen;
    UInt8		notif;

    if (!me->fReady)
	{
		return;
	}
	if (rc == kIOReturnSuccess)	// If operation returned ok
    {
        dLen = COMM_BUFF_SIZE - remaining;
        ELG(rc, dLen, 'cRC+', "org_lucidcake_driver_AJZaurusUSB::commReadComplete");
		
            // Now look at the state stuff
            
        LogData(kUSBAnyDirn, dLen, me->fCommPipeBuffer);
        
        notif = me->fCommPipeBuffer[1];
        if (dLen > 7)
        {
            switch(notif)
            {
                case kNetwork_Connection:
                    me->fLinkStatus = me->fCommPipeBuffer[2];
                    ELG(0, me->fLinkStatus, 'cRNC', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - kNetwork_Connection");
                    break;
                case kConnection_Speed_Change:				// In you-know-whose format
                    me->fUpSpeed = USBToHostLong((UInt32)me->fCommPipeBuffer[8]);
                    me->fDownSpeed = USBToHostLong((UInt32)me->fCommPipeBuffer[13]);
                    ELG(me->fUpSpeed, me->fDownSpeed, 'cRCS', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - kConnection_Speed_Change");
                    break;
                default:
                    ALERT(0, notif, 'cRUn', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - Unknown notification");
                    break;
            }
        } else {
            ALERT(0, notif, 'cRIn', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - Invalid notification");
        }
    } else {
        ALERT(0, rc, 'cRC-', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - IO error");
    }

        // Queue the next read, only if not aborted
	
    if (rc != kIOReturnAborted)
    {
        ior = me->fCommPipe->Read(me->fCommPipeMDP, &me->fCommCompletionInfo, NULL);
        if (ior != kIOReturnSuccess)
        {
            ALERT(0, ior, 'cRF-', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - Failed to queue next read");
            if (ior == kIOUSBPipeStalled)
            {
                me->fCommPipe->Reset();
                ior = me->fCommPipe->Read(me->fCommPipeMDP, &me->fCommCompletionInfo, NULL);
                if (ior != kIOReturnSuccess)
                {
                    ALERT(0, ior, 'cR--', "org_lucidcake_driver_AJZaurusUSB::commReadComplete - Failed, read dead");
                    me->fCommDead = true;
                }
            }
        }
    }
    
    return;
	
}/* end commReadComplete */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::dataReadComplete
//
//		Inputs:		obj - me
//				param - unused
//				rc - return code
//				remaining - what's left
//
//		Outputs:	None
//
//		Desc:		BulkIn pipe (Data interface) read completion routine
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::dataReadComplete(void *obj, void *param, IOReturn rc, UInt32 remaining)
{
    org_lucidcake_driver_AJZaurusUSB	*me = (org_lucidcake_driver_AJZaurusUSB*)obj;
    IOReturn		ior;

    if (!me->fReady)
	{
		return;
	}
	if (rc == kIOReturnSuccess)	// If operation returned ok
    {	
        ELG(0, me->fMax_Block_Size - remaining, 'dRC+', "dataReadComplete");
		
        LogData(kUSBIn, (me->fMax_Block_Size - remaining), me->fPipeInBuffer);
	
            // Move the incoming bytes up the stack

        me->receivePacket(me->fPipeInBuffer, me->fMax_Block_Size - remaining);
	
    } else {
        ELG(0, rc, 'dRc-', "org_lucidcake_driver_AJZaurusUSB::dataReadComplete - Read completion io err");
        if (rc == kIOUSBPipeStalled)
        {
            rc = me->clearPipeStall(me->fInPipe);
            if (rc != kIOReturnSuccess)
            {
                ALERT(0, rc, 'dR--', "org_lucidcake_driver_AJZaurusUSB::dataReadComplete - clear stall failed");
				return;
            }
        } else {
			return;
		}
    }
    
        // Queue the next read
	
	ior = me->fInPipe->Read(me->fPipeInMDP, &me->fReadCompletionInfo, NULL);
	if (ior != kIOReturnSuccess)
	{
		ALERT(0, ior, 'dRe-', "org_lucidcake_driver_AJZaurusUSB::dataReadComplete - Failed to queue read");
		if (ior == kIOUSBPipeStalled)
		{
			me->fInPipe->Reset();
			ior = me->fInPipe->Read(me->fPipeInMDP, &me->fReadCompletionInfo, NULL);
			if (ior != kIOReturnSuccess)
			{
				ALERT(0, ior, 'dR--', "org_lucidcake_driver_AJZaurusUSB::dataReadComplete - Failed, read dead");
				me->fDataDead = true;
			}
		}
	}

    return;
	
}/* end dataReadComplete */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::dataWriteComplete
//
//		Inputs:		obj - me
//				param - pool index
//				rc - return code
//				remaining - what's left
//
//		Outputs:	None
//
//		Desc:		BulkOut pipe (Data interface) write completion routine
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::dataWriteComplete(void *obj, void *param, IOReturn rc, UInt32 remaining)
{
    org_lucidcake_driver_AJZaurusUSB	*me = (org_lucidcake_driver_AJZaurusUSB *)obj;
    //struct mbuf		*m;
    //UInt32		pktLen = 0;
    //UInt32		numbufs = 0;
    UInt32		poolIndx;
	SInt32		dataCount;
	bool		stalled;

	if (!me->fReady)
	{
		return;
	}
    poolIndx = (UInt32)param;
    
    if (rc == kIOReturnSuccess)						// If operation returned ok
    {	
        ELG(rc, poolIndx, 'dWC+', "org_lucidcake_driver_AJZaurusUSB::dataWriteComplete");
		IOSimpleLockLock(me->fLock);
		--(me->fDataCount);
		dataCount = me->fDataCount;
		stalled = me->fOutputStalled;
		me->fOutputStalled = false;
        if (me->fPipeOutBuff[poolIndx].m != NULL)			// Null means zero length write
        {
            //m = me->fPipeOutBuff[poolIndx].m;
            //while (m)
            //{
            //    pktLen += m->m_len;
            //    numbufs++;
            //    m = m->m_next;
            //}
            
            //me->freePacket(me->fPipeOutBuff[poolIndx].m);		// Free the mbuf
			// AJ: The above is commented out since the mbuf is now freed in the
			// USBTransmitPacket method
            me->fPipeOutBuff[poolIndx].m = NULL;
			
            //if ((pktLen % me->fOutPacketSize) == 0)			// If it was a multiple of max packet size then we need to do a zero length write
            //{
            //    ELG(rc, pktLen, 'dWCz', "com_apple_driver_dts_USBCDCEthernet::dataWriteComplete - writing zero length packet");
            //    me->fPipeOutBuff[poolIndx].pipeOutMDP->setLength(0);
            //    me->fPipeOutBuff[poolIndx].writeCompletionInfo.parameter = NULL;
            //    me->fOutPipe->Write(me->fPipeOutBuff[poolIndx].pipeOutMDP, &(me->fPipeOutBuff[poolIndx].writeCompletionInfo));
            //}
			
			// AJ: Because of the padding protocol used by the Zaurus, the above never happens
			// (and I'm not sure what it was good for in the first place).
        }
		IOSimpleLockUnlock(me->fLock);
    } else {
        ELG(0, rc, 'dWe-', "org_lucidcake_driver_AJZaurusUSB::dataWriteComplete - IO err");
        if (rc != kIOReturnAborted)
        {
            rc = me->clearPipeStall(me->fOutPipe);
            if (rc != kIOReturnSuccess)
            {
                ALERT(0, rc, 'dW--', "org_lucidcake_driver_AJZaurusUSB::dataWriteComplete - clear stall failed (trying to continue)");
            }
        }
    }
	//OSIncrementAtomic(&me->fSyncCount);
	//ALERT(me->fSyncCount, 0, 'syn+', "org_lucidcake_driver_AJZaurusUSB::dataWriteComplete - ++Sync");
	if (stalled && dataCount == 0) {
		//ALERT(0, 0, '----', "service");
		me->fTransmitQueue->service(IOBasicOutputQueue::kServiceAsync);	// revive a stalled queue
	}
	me->fTransmitQueue->service();	// revive a stalled queue

	return;
	
}/* end dataWriteComplete */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::merWriteComplete
//
//		Inputs:		obj - me
//				param - parameter block (may or may not be present depending on request) 
//				rc - return code
//				remaining - what's left
//
//		Outputs:	None
//
//		Desc:		Management element request write completion routine
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::merWriteComplete(void *obj, void *param, IOReturn rc, UInt32 remaining)
{
    IOUSBDevRequest	*MER = (IOUSBDevRequest*)param;
    UInt16		dataLen;
	
    if (MER)
    {
        if (rc == kIOReturnSuccess)
        {
            ELG(MER->bRequest, remaining, 'mWC+', "org_lucidcake_driver_AJZaurusUSB::merWriteComplete");
        } else {
            ALERT(MER->bRequest, rc, 'mWC-', "org_lucidcake_driver_AJZaurusUSB::merWriteComplete - io err");
        }
		
        dataLen = MER->wLength;
        ELG(0, dataLen, 'mWC ', "org_lucidcake_driver_AJZaurusUSB::merWriteComplete - data length");
        if ((dataLen != 0) && (MER->pData))
        {
            IOFree(MER->pData, dataLen);
        }
        IOFree(MER, sizeof(IOUSBDevRequest));
		
    } else {
        if (rc == kIOReturnSuccess)
        {
            ELG(0, remaining, 'mWr+', "org_lucidcake_driver_AJZaurusUSB::merWriteComplete (request unknown)");
        } else {
            ALERT(0, rc, 'rWr-', "org_lucidcake_driver_AJZaurusUSB::merWriteComplete (request unknown) - io err");
        }
    }
    return;
	
}/* end merWriteComplete */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::statsWriteComplete
//
//		Inputs:		obj - me
//				param - parameter block 
//				rc - return code
//				remaining - what's left
//
//		Outputs:	None
//
//		Desc:		Ethernet statistics request write completion routine
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::statsWriteComplete(void *obj, void *param, IOReturn rc, UInt32 remaining)
{
    org_lucidcake_driver_AJZaurusUSB	*me = (org_lucidcake_driver_AJZaurusUSB *)obj;
    IOUSBDevRequest	*STREQ = (IOUSBDevRequest*)param;
    UInt16		currStat;
	
    if (STREQ)
    {
        if (rc == kIOReturnSuccess && me->fReady)
        {
            ELG(STREQ->bRequest, remaining, 'sWC+', "org_lucidcake_driver_AJZaurusUSB::statsWriteComplete");
            currStat = STREQ->wValue;
            switch(currStat)
            {
                case kXMIT_OK_REQ:
                    me->fpNetStats->outputPackets = USBToHostLong(me->fStatValue);
                    break;
                case kRCV_OK_REQ:
                    me->fpNetStats->inputPackets = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_ERROR_REQ:
                    me->fpNetStats->outputErrors = USBToHostLong(me->fStatValue);
                    break;
                case kRCV_ERROR_REQ:
                    me->fpNetStats->inputErrors = USBToHostLong(me->fStatValue);
                    break;
                case kRCV_CRC_ERROR_REQ:
                    me->fpEtherStats->dot3StatsEntry.fcsErrors = USBToHostLong(me->fStatValue); 
                    break;
                case kRCV_ERROR_ALIGNMENT_REQ:
                    me->fpEtherStats->dot3StatsEntry.alignmentErrors = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_ONE_COLLISION_REQ:
                    me->fpEtherStats->dot3StatsEntry.singleCollisionFrames = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_MORE_COLLISIONS_REQ:
                    me->fpEtherStats->dot3StatsEntry.multipleCollisionFrames = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_DEFERRED_REQ:
                    me->fpEtherStats->dot3StatsEntry.deferredTransmissions = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_MAX_COLLISION_REQ:
                    me->fpNetStats->collisions = USBToHostLong(me->fStatValue);
                    break;
                case kRCV_OVERRUN_REQ:
                    me->fpEtherStats->dot3StatsEntry.frameTooLongs = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_TIMES_CARRIER_LOST_REQ:
                    me->fpEtherStats->dot3StatsEntry.carrierSenseErrors = USBToHostLong(me->fStatValue);
                    break;
                case kXMIT_LATE_COLLISIONS_REQ:
                    me->fpEtherStats->dot3StatsEntry.lateCollisions = USBToHostLong(me->fStatValue);
                    break;
                default:
                    ALERT(currStat, rc, 'sWI-', "org_lucidcake_driver_AJZaurusUSB::statsWriteComplete - Invalid stats code");
                    break;
            }

        } else {
            ALERT(STREQ->bRequest, rc, 'sWC-', "org_lucidcake_driver_AJZaurusUSB::statsWriteComplete - io err");
        }
		
        IOFree(STREQ, sizeof(IOUSBDevRequest));
    } else {
        if (rc == kIOReturnSuccess)
        {
            ELG(0, remaining, 'sWr+', "org_lucidcake_driver_AJZaurusUSB::statsWriteComplete (request unknown)");
        } else {
            ALERT(0, rc, 'sWr-', "org_lucidcake_driver_AJZaurusUSB::statsWriteComplete (request unknown) - io err");
        }
    }
	
    me->fStatValue = 0;
    me->fStatInProgress = false;
    return;
	
}/* end statsWriteComplete */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::init
//
//		Inputs:		properties - data (keys and values) used to match
//
//		Outputs:	Return code - true (init successful), false (init failed)
//
//		Desc:		Initialize the driver.
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::init(OSDictionary *properties)
{
    UInt32	i;

    g.evLogBufp = NULL;
        
#if USE_ELG
    AllocateEventLog(kEvLogSize);
    ELG(&g, g.evLogBufp, 'USBM', "org_lucidcake_driver_AJZaurusUSB::init - event logging set up.");

    waitForService(resourceMatching("kdp"));
#endif /* USE_ELG */

    ELG(0, 0, 'init', "org_lucidcake_driver_AJZaurusUSB::init");
    
    if (super::init(properties) == false)
    {
        ALERT(0, 0, 'in--', "org_lucidcake_driver_AJZaurusUSB::init - initialize super failed");
        return false;
    }
    
        // Set some defaults
    
    fMax_Block_Size = MAX_BLOCK_SIZE;
    fCurrStat = 0;
    fStatInProgress = false;
    fDataDead = false;
    fCommDead = false;
    fPacketFilter = kPACKET_TYPE_DIRECTED | kPACKET_TYPE_BROADCAST | kPACKET_TYPE_MULTICAST;
	// AJ: use padding (TODO: find out whether this is really necessary)
	// -->
	fPadded = true;
	fChecksum = true;
	// <--
    
    for (i=0; i<kOutBufPool; i++)
    {
        fPipeOutBuff[i].pipeOutMDP = NULL;
        fPipeOutBuff[i].pipeOutBuffer = NULL;
        fPipeOutBuff[i].m = NULL;
    }

	fLock = IOSimpleLockAlloc();

    return true;

}/* end init*/

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::start
//
//		Inputs:		provider - my provider
//
//		Outputs:	Return code - true (it's me), false (sorry it probably was me, but I can't configure it)
//
//		Desc:		This is called once it has beed determined I'm probably the best 
//				driver for this device.
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::start(IOService *provider)
{
    UInt8	configs;	// number of device configurations

    ELG(this, provider, 'strt', "org_lucidcake_driver_AJZaurusUSB::start - this, provider.");
    if(!super::start(provider))
    {
        ALERT(0, 0, 'SS--', "org_lucidcake_driver_AJZaurusUSB::start - start super failed");
        return false;
    }

	// Get my USB device provider - the device

    fpDevice = OSDynamicCast(IOUSBDevice, provider);
    if(!fpDevice)
    {
        ALERT(0, 0, 'Dev-', "org_lucidcake_driver_AJZaurusUSB::start - provider invalid");
        stop(provider);
        return false;
    }

	// Let's see if we have any configurations to play with
		
    configs = fpDevice->GetNumConfigurations();
    if (configs < 1)
    {
        ALERT(0, 0, 'Cfg-', "org_lucidcake_driver_AJZaurusUSB::start - no configurations");
        stop(provider);
        return false;
    }
	
	// Now take control of the device and configure it
		
    if (!fpDevice->open(this))
    {
        ALERT(0, 0, 'Opn-', "org_lucidcake_driver_AJZaurusUSB::start - unable to open device");
        stop(provider);
        return false;
    }
	
    if (!configureDevice(configs))
    {
        ALERT(0, 0, 'Nub-', "org_lucidcake_driver_AJZaurusUSB::start - failed");
        fpDevice->close(this);
        fpDevice = NULL;
        stop(provider);
        return false;
    }
    
    ELG(0, 0, 'Nub+', "org_lucidcake_driver_AJZaurusUSB::start - successful");
    
    return true;
    	
}/* end start */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::free
//
//		Inputs:		None
//
//		Outputs:	None
//
//		Desc:		Clean up and free the log 
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::free()
{

    ELG(0, 0, 'free', "org_lucidcake_driver_AJZaurusUSB::free");
	
#if USE_ELG
    if (g.evLogBuf)
    	IOFree(g.evLogBuf, kEvLogSize);
#endif /* USE_ELG */

    super::free();
	IOSimpleLockFree(fLock);
    return;
	
}/* end free */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::stop
//
//		Inputs:		provider - my provider
//
//		Outputs:	None
//
//		Desc:		Stops
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::stop(IOService *provider)
{
    
    ELG(0, 0, 'stop', "org_lucidcake_driver_AJZaurusUSB::stop");
    
    if (fNetworkInterface)
    {
		detachInterface(fNetworkInterface);
        fNetworkInterface->release();
        fNetworkInterface = NULL;
    }

    
    if (fCommInterface)	
    {
		if (!fSingleInterface) {
			fCommInterface->close(this);
		}
		fCommInterface->release();
		fCommInterface = NULL;	
    }
	
    if (fDataInterface)	
    { 
        fDataInterface->close(this);	
        fDataInterface->release();
        fDataInterface = NULL;	
    }

    if (fpDevice)
    {
        fpDevice->close(this);
        fpDevice = NULL;
    }
    
    if (fMediumDict)
    {
        fMediumDict->release();
        fMediumDict = NULL;
    }
    
	if (fTransmitQueue)
	{
		fTransmitQueue->release();
		fTransmitQueue = NULL;
	}
	
    super::stop(provider);
    
    return;
	
}/* end stop */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::configureDevice
//
//		Inputs:		numConfigs - number of configurations present
//
//		Outputs:	return Code - true (device configured), false (device not configured)
//
//		Desc:		Finds the configurations and then the appropriate interfaces etc.
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::configureDevice(UInt8 numConfigs)
{
    IOUSBFindInterfaceRequest		req;			// device request
    const IOUSBInterfaceDescriptor	*altInterfaceDesc;
    IOReturn				ior = kIOReturnSuccess;
    UInt16				numends = 0;
    UInt16				alt;
    bool				goodCall;
       
    ELG(0, numConfigs, 'cDev', "org_lucidcake_driver_AJZaurusUSB::configureDevice");
    	
        // Initialize and "configure" the device
        
    if (!initDevice(numConfigs))
    {
        ALERT(0, 0, 'cDi-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - initDevice failed");
        return false;
    }

        // Get the Comm. Class interface

	req.bInterfaceClass	= fInterfaceClass;
	if (fSingleInterface) {
		req.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
	} else {
		req.bInterfaceSubClass = kEthernetControlModel;
	}
    req.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
    req.bAlternateSetting = kIOUSBFindInterfaceDontCare;
    
    fCommInterface = fpDevice->FindNextInterface(NULL, &req);
    if (!fCommInterface)
    {
        ALERT(0, 0, 'FIC-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - Finding the first CDC interface failed");
        return false;
    }
    
    if (!getFunctionalDescriptors())
    {
        ALERT(0, 0, 'cDi-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - getFunctionalDescriptors failed");
        return false;
    }

	if (!fSingleInterface) {
		// AJ: If we have an MDLM device, comm and data interface are identical,
		// so only open the comm interface here if we DON'T hava an MDLM device.
		goodCall = fCommInterface->open(this);
		if (!goodCall)
		{
			ALERT(0, 0, 'epC-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - open comm interface failed.");
			fCommInterface = NULL;
			return false;
		}
    }
	
    fCommInterfaceNumber = fCommInterface->GetInterfaceNumber();
    
        // Now get the Data Class interface
	
	if (fSingleInterface) {
		fDataInterface = fCommInterface;
	} else {
		req.bInterfaceClass = kUSBDataClass;
		req.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
		req.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
		req.bAlternateSetting = kIOUSBFindInterfaceDontCare;
    
		fDataInterface = fpDevice->FindNextInterface(NULL, &req);    
	}
	
    if (fDataInterface)
    {
        numends = fDataInterface->GetNumEndpoints();
        if (numends > 1)					// There must be (at least) two bulk endpoints
        {
            //ALERT(0, 0, '----', "Standard interface works");
            ELG(numends, fDataInterface, 'cDD+', "org_lucidcake_driver_AJZaurusUSB::configureDevice - Data Class interface found");
            goodCall = fDataInterface->open(this);
            if (!goodCall)
			{
                ALERT(0, 0, 'cDD-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - open data interface failed.");
                numends = 0;
            }
        } else {
            //ALERT(0, 0, '----', "Standard interface DOESN'T work");
            altInterfaceDesc = fDataInterface->FindNextAltInterface(NULL, &req);
            if (!altInterfaceDesc)
            {
                ALERT(0, 0, 'cDn-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - FindNextAltInterface failed");
            }
            while (altInterfaceDesc)
            {
                numends = altInterfaceDesc->bNumEndpoints;
                if (numends > 1)
                {
                    goodCall = fDataInterface->open(this);
                    if (goodCall)
                    {
                        alt = altInterfaceDesc->bAlternateSetting;
                        ELG(numends, alt, 'cD++', "org_lucidcake_driver_AJZaurusUSB::configureDevice - Data Class interface (alternate) found");
                        ior = fDataInterface->SetAlternateInterface(this, alt);
                        if (ior == kIOReturnSuccess)
                        {
                            ELG(0, 0, 'cDA+', "org_lucidcake_driver_AJZaurusUSB::configureDevice - Alternate set");
                            break;
                        } else {
                            ALERT(0, 0, 'cDS-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - SetAlternateInterface failed");
                            numends = 0;
                        }
                    } else {
                        ALERT(0, 0, 'cDD-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - open data interface failed.");
                        numends = 0;
                    }
                } else {
                    ELG(0, altInterfaceDesc, 'cDe-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - No endpoints this alternate");
                }
                altInterfaceDesc = fDataInterface->FindNextAltInterface(altInterfaceDesc, &req);
            }
        }
    } else {
        ALERT(0, 0, 'cDr-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - FindNextInterface failed");
    }

    if (numends < 2)
    {
        ALERT(0, 0, 'cDs-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - Finding a Data Class interface failed");
		if (!fSingleInterface) {
			fCommInterface->close(this);
		}
        fCommInterface = NULL;
        return false;
    } else {
    
        fCommInterface->retain();
        fDataInterface->retain();

            // Found both so now let's publish the interface
	
        if (!createNetworkInterface())
        {
            ALERT(0, 0, 'cDc-', "org_lucidcake_driver_AJZaurusUSB::configureDevice - createNetworkInterface failed");
            fCommInterface->release();
			if (!fSingleInterface) {
				fCommInterface->close(this);
			}
            fCommInterface = NULL;
            fDataInterface->release();
            fDataInterface->close(this);
            fDataInterface = NULL;
            return false;
        }
    }
    
	//ALERT(fbmAttributes, 0, 'cDAt', "org_lucidcake_driver_AJZaurusUSB::configureDevice - attributes from configuration descriptor");
    /*if (fbmAttributes & kUSBAtrBusPowered)
    {
        ior = fpDevice->SuspendDevice(true);         // Suspend the device (if supported and bus powered)
        if (ior)
        {
            ALERT(0, ior, 'cCSD', "org_lucidcake_driver_AJZaurusUSB::configureDevice - SuspendDevice error");
        }
    }*/
	
	// AJ: The above is commented out because it caused a lot of trouble, and I
	// can't imagine why it would make sense to *suspend* the device here.

    return true;

}/* end configureDevice */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::initDevice
//
//		Inputs:		numConfigs - number of configurations present
//
//		Outputs:	return Code - true (CDC present), false (CDC not present)
//
//		Desc:		Determines if this is a CDC compliant device and then sets the configuration
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::dumpDevice(UInt8 numConfigs)
{
    IOUSBFindInterfaceRequest		req;
    const IOUSBConfigurationDescriptor	*cd = NULL;		// configuration descriptor
    IOUSBInterfaceDescriptor 		*intf = NULL;		// interface descriptor
    IOReturn				ior = kIOReturnSuccess;
    UInt8				cval;
	UInt8				intfCount = 0;
       
	ALERT(0, numConfigs, '----', "Enumerating configurations ...");
	
    for (cval=0; cval<numConfigs; cval++)
    {
		ALERT(0, cval, '----', "  Config found");
     	cd = fpDevice->GetFullConfigurationDescriptor(cval);
     	if (!cd)
    	{
            ALERT(0, cval, '----', "  Enumeration failure");
        } else {
			intf = NULL;
			intfCount = 0;
			req.bInterfaceClass	= kIOUSBFindInterfaceDontCare;
			req.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
			req.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
			req.bAlternateSetting = kIOUSBFindInterfaceDontCare;
			do {
				ior = fpDevice->FindNextInterfaceDescriptor(cd, intf, &req, &intf);
				if (ior == kIOReturnSuccess) {
					ALERT(intfCount, cval, '----', "    Interface found");
					ALERT(intf->bLength, intf->bDescriptorType, '----', "      bLength, bDescriptorType");
					ALERT(intf->bInterfaceNumber, intf->bAlternateSetting, '----', "      bInterfaceNumber, bAlternateSetting");
					ALERT(intf->bNumEndpoints, 0, '----', "      bNumEndpoints");
					ALERT(intf->bInterfaceClass, intf->bInterfaceSubClass, '----', "      bInterfaceClass, bInterfaceSubClass");
					ALERT(intf->bInterfaceProtocol, intf->iInterface, '----', "      bInterfaceProtocol, iInterface");
					++intfCount;
				}
			} while (ior == kIOReturnSuccess);
        }
    }
}
	
bool org_lucidcake_driver_AJZaurusUSB::initDevice(UInt8 numConfigs)
{
    IOUSBFindInterfaceRequest		req;
    const IOUSBConfigurationDescriptor	*cd = NULL;		// configuration descriptor
    IOUSBInterfaceDescriptor 		*intf = NULL;		// interface descriptor
    IOReturn				ior = kIOReturnSuccess;
    UInt8				cval;
    UInt8				config = 0;
    bool				goodconfig = false;
       
	dumpDevice(numConfigs);
    ELG(0, numConfigs, 'cDev', "org_lucidcake_driver_AJZaurusUSB::initDevice");
    	
        // Make sure we have a CDC interface to play with
        
    for (cval=0; cval<numConfigs; cval++)
    {
    	ELG(0, cval, 'CkCn', "org_lucidcake_driver_AJZaurusUSB::initDevice - Checking Configuration");
		
     	cd = fpDevice->GetFullConfigurationDescriptor(cval);
     	if (!cd)
    	{
            ALERT(0, 0, 'GFC-', "org_lucidcake_driver_AJZaurusUSB::initDevice - Error getting the full configuration descriptor");
        } else {
            req.bInterfaceClass	= kUSBCommClass;
            //req.bInterfaceSubClass = kEthernetControlModel;
            req.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
            req.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
            req.bAlternateSetting = kIOUSBFindInterfaceDontCare;
            ior = fpDevice->FindNextInterfaceDescriptor(cd, intf, &req, &intf);
			if (ior != kIOReturnSuccess)
			{
				// try again - this time with vendor specific class
				// (used by Opie on the iPAQ)
				req.bInterfaceClass = kUSBVendorSpecificClass;
				ior = fpDevice->FindNextInterfaceDescriptor(cd, intf, &req, &intf);
			}
			if (ior == kIOReturnSuccess)
			{
				ALERT(intf->bInterfaceClass, intf->bInterfaceSubClass, 'A300', "org_lucidcake_driver_AJZaurusUSB::initDevice - Interface class and sub class");
				if (intf->bInterfaceSubClass == kMDLM) {
					fSingleInterface = true;
				} else {
					fSingleInterface = false;
				}
				fInterfaceClass = intf->bInterfaceClass;
				if (fInterfaceClass == kUSBVendorSpecificClass) {
					fSingleInterface = true;
					// for now: no padding and/or CRC
					fPadded = false;
					fChecksum = false;
				}
                if (intf)
                {
                    config = cd->bConfigurationValue;
                    ELG(0, config, 'FNI+', "org_lucidcake_driver_AJZaurusUSB::initDevice - Interface descriptor found");
                    goodconfig = true;					// We have at least one CDC interface in this configuration
                    break;
                } else {
                    ALERT(0, config, 'FNI-', "org_lucidcake_driver_AJZaurusUSB::initDevice - That's weird the interface was null");
                }
            } else {
                ALERT(ior, cval, 'FNID', "org_lucidcake_driver_AJZaurusUSB::initDevice - No CDC interface found this configuration");
            }
        }
    }
        
    if (goodconfig)
    {
        ior = fpDevice->SetConfiguration(this, config);
        if (ior != kIOReturnSuccess)
        {
            ALERT(0, ior, 'SCo-', "org_lucidcake_driver_AJZaurusUSB::initDevice - SetConfiguration error");
            goodconfig = false;			
        }
    } else {
        return false;
    }
    
    fbmAttributes = cd->bmAttributes;
    ELG(fbmAttributes, kUSBAtrRemoteWakeup, 'GFbA', "org_lucidcake_driver_AJZaurusUSB::initDevice - Configuration bmAttributes");
    
        // Save the ID's
    
    fVendorID = fpDevice->GetVendorID();
    fProductID = fpDevice->GetProductID();
    
    return goodconfig;

}/* end initDevice */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors
//
//		Inputs:		
//
//		Outputs:	return - true (descriptors ok), false (somethings not right or not supported)	
//
//		Desc:		Finds all the functional descriptors for the specific interface
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors()
{
    bool				gotDescriptors = false;
    bool				configok = true;
    bool				enet = false;
    IOReturn				ior;
    const HeaderFunctionalDescriptor 	*funcDesc = NULL;
    EnetFunctionalDescriptor		*ENETFDesc;
       
    ELG(0, 0, 'gFDs', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors");
        
    do
    {
        (IOUSBDescriptorHeader*)funcDesc = fCommInterface->FindNextAssociatedDescriptor((void*)funcDesc, CS_INTERFACE);
        if (!funcDesc)
        {
            gotDescriptors = true;
        } else {
            switch (funcDesc->bDescriptorSubtype)
            {
                case Header_FunctionalDescriptor:
                    ELG(funcDesc->bDescriptorType, funcDesc->bDescriptorSubtype, 'gFHd', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - Header Functional Descriptor");
                    break;
                case Enet_Functional_Descriptor:
                    (const HeaderFunctionalDescriptor *)ENETFDesc = funcDesc;
                    ELG(funcDesc->bDescriptorType, funcDesc->bDescriptorSubtype, 'gFEN', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - Ethernet Functional Descriptor");
                    enet = true;
                    break;
                case Union_FunctionalDescriptor:
                    ELG(funcDesc->bDescriptorType, funcDesc->bDescriptorSubtype, 'gFUn', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - Union Functional Descriptor");
                    break;
                default:
                    ELG(funcDesc->bDescriptorType, funcDesc->bDescriptorSubtype, 'gFFD', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - unknown Functional Descriptor");
                    break;
            }
        }
    } while(!gotDescriptors);

    if (!enet)
    {
        configok = false;					// The Enet Func. Desc.  must be present
    } else {
    
            // Determine who is collecting the input/output network stats.
    
        if (!(ENETFDesc->bmEthernetStatistics[0] & kXMIT_OK))
        {
            fOutputPktsOK = true;
        } else {
            fOutputPktsOK = false;
        }
        if (!(ENETFDesc->bmEthernetStatistics[0] & kRCV_OK))
        {
            fInputPktsOK = true;
        } else {
            fInputPktsOK = false;
        }
        if (!(ENETFDesc->bmEthernetStatistics[0] & kXMIT_ERROR))
        {
            fOutputErrsOK = true;
        } else {
            fOutputErrsOK = false;
        }
        if (!(ENETFDesc->bmEthernetStatistics[0] & kRCV_ERROR))
        {
            fInputErrsOK = true;
        } else {
            fInputErrsOK = false;
        }
        
            // Save the stats (it's bit mapped)
        
        fEthernetStatistics[0] = ENETFDesc->bmEthernetStatistics[0];
        fEthernetStatistics[1] = ENETFDesc->bmEthernetStatistics[1];
        fEthernetStatistics[2] = ENETFDesc->bmEthernetStatistics[2];
        fEthernetStatistics[3] = ENETFDesc->bmEthernetStatistics[3];
        
            // Save the multicast filters (remember it's intel format)
        
        fMcFilters = USBToHostWord((UInt16)ENETFDesc->wNumberMCFilters[0]);
        
            // Get the Ethernet address
    
        if (ENETFDesc->iMACAddress != 0)
        {	
            ior = fpDevice->GetStringDescriptor(ENETFDesc->iMACAddress, (char *)&fEaddr, 6);
            if (ior == kIOReturnSuccess)
            {
#if LOG_DATA
                ELG(0, ENETFDesc->iMACAddress, 'gFEA', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - Ethernet address");
#endif
                LogData(kUSBAnyDirn, 6, fEaddr);
                
                fMax_Block_Size = USBToHostWord((UInt16)ENETFDesc->wMaxSegmentSize[0]);
                ELG(0, fMax_Block_Size, 'gFMs', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - Maximum segment size");
            } else {
                ALERT(0, 0, 'gFAe', "org_lucidcake_driver_AJZaurusUSB::getFunctionalDescriptors - Error retrieving Ethernet address");
                configok = false;
            }
        } else {
            configok = false;
        }
    }

	// Use some defaults if all else failed.
	// FIXME: This is a hack! An evil one!
	// (I really need to find out how to interpret the descriptors returned by the iPaq.)
	if (!configok)
	{
		fOutputPktsOK = false;
		fInputPktsOK = false;
		fOutputErrsOK = false;
		fInputErrsOK = false;
		fEthernetStatistics[0] = 0;
		fEthernetStatistics[1] = 0;
		fEthernetStatistics[2] = 0;
		fEthernetStatistics[3] = 0;
		fMcFilters = 0;
		fMax_Block_Size = 0xea00; //MAX_BLOCK_SIZE;
		fEaddr[0] = 0x34;
		fEaddr[1] = 0x30;
		fEaddr[2] = 0x30;
		fEaddr[3] = 0x30;
		fEaddr[4] = 0x30;
		fEaddr[5] = 0x00;
		configok = true;
	}
    
    return configok;
    
}/* end getFunctionalDescriptors */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::createNetworkInterface
//
//		Inputs:		
//
//		Outputs:	return Code - true (created and initialilzed ok), false (it failed)
//
//		Desc:		Creates and initializes the network interface
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::createNetworkInterface()
{
	
    ELG(0, 0, 'crIf', "org_lucidcake_driver_AJZaurusUSB::createNetworkInterface");
    
            // Allocate memory for buffers etc

    fTransmitQueue = (IOBasicOutputQueue*)getOutputQueue();
    if (!fTransmitQueue) 
    {
        ALERT(0, 0, 'crO-', "org_lucidcake_driver_AJZaurusUSB::createNetworkInterface - Output queue initialization failed");
        return false;
    }
    fTransmitQueue->retain();
    
        // Get a reference to the IOWorkLoop in our superclass
        
    fWorkLoop = getWorkLoop();
    
        // Allocate Timer event source
        
    fTimerSource = IOTimerEventSource::timerEventSource(this, timerFired);
    if (fTimerSource == NULL)
    {
        ALERT(0, 0, 'crT-', "org_lucidcake_driver_AJZaurusUSB::createNetworkInterface - Allocate Timer event source failed");
        return false;
    }
    
    if (fWorkLoop->addEventSource(fTimerSource) != kIOReturnSuccess)
    {
        ALERT(0, 0, 'crt-', "org_lucidcake_driver_AJZaurusUSB::start - Add Timer event source failed");        
        return false;
    }

        // Attach an IOEthernetInterface client
        
    ELG(0, 0, 'crai', "org_lucidcake_driver_AJZaurusUSB::createNetworkInterface - attaching and registering interface");
    
    if (!attachInterface((IONetworkInterface **)&fNetworkInterface, true))
    {	
        ALERT( 0, 0, 'crt-', "org_lucidcake_driver_AJZaurusUSB::createNetworkInterface - attachInterface failed");      
        return false;
    }
    
        // Ready to service interface requests
    
    //fNetworkInterface->registerService();
	// AJ: this is already achieved by attachInterface
	// with a second parameter of true.
    
    ELG(0, 0, 'crEx', "org_lucidcake_driver_AJZaurusUSB::createNetworkInterface - Exiting, successful");

    return true;
	
}/* end createNetworkInterface */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::enable
//
//		Inputs:		netif - the interface being enabled
//
//		Outputs:	Return code - kIOReturnSuccess or kIOReturnIOError
//
//		Desc:		Called by IOEthernetInterface client to enable the controller.
//				This method is always called while running on the default workloop
//				thread
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::enable(IONetworkInterface *netif)
{
    IONetworkMedium	*medium;
    IOMediumType    	mediumType = kIOMediumEthernet10BaseT | kIOMediumOptionFullDuplex;
    
    ELG(0, netif, 'enbl', "org_lucidcake_driver_AJZaurusUSB::enable");

        // If an interface client has previously enabled us,
        // and we know there can only be one interface client
        // for this driver, then simply return success.

    if (fNetifEnabled)
    {
        ELG(0, 0, 'enae', "org_lucidcake_driver_AJZaurusUSB::enable - already enabled");
        return kIOReturnSuccess;
    }
    
    if ((fReady == false) && !wakeUp())
    {
        ALERT(0, fReady, 'enr-', "org_lucidcake_driver_AJZaurusUSB::enable - failed");
        return kIOReturnIOError;
    }

        // Mark the controller as enabled by the interface.

    fNetifEnabled = true;
    
        // Assume an active link (leave this in for now - until we know better)
        // Should probably use the values returned in the Network Connection notification
        // that is if we have an interrupt pipe, otherwise default to these
    
    fLinkStatus = 1;
    
    medium = IONetworkMedium::getMediumWithType(fMediumDict, mediumType);
    ELG(mediumType, medium, 'enam', "org_lucidcake_driver_AJZaurusUSB::enable - medium type and pointer");
    setLinkStatus(kIONetworkLinkActive | kIONetworkLinkValid, medium, 10 * 1000000);
    ELG(0, 0, 'enaL', "org_lucidcake_driver_AJZaurusUSB::enable - LinkStatus set");
    
        // Start our IOOutputQueue object.

    fTransmitQueue->setCapacity(TRANSMIT_QUEUE_SIZE);
    ELG(0, TRANSMIT_QUEUE_SIZE, 'enaC', "org_lucidcake_driver_AJZaurusUSB::enable - capicity set");
    fTransmitQueue->start();
    ELG(0, 0, 'enaT', "org_lucidcake_driver_AJZaurusUSB::enable - transmit queue started");
    
    USBSetPacketFilter();
    ELG(0, 0, 'enaP', "org_lucidcake_driver_AJZaurusUSB::enable - packet filter applied");

    return kIOReturnSuccess;
    
}/* end enable */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::disable
//
//		Inputs:		netif - the interface being disabled
//
//		Outputs:	Return code - kIOReturnSuccess
//
//		Desc:		Called by IOEthernetInterface client to disable the controller.
//				This method is always called while running on the default workloop
//				thread
//
/****************************************************************************************************/
 
IOReturn org_lucidcake_driver_AJZaurusUSB::disable(IONetworkInterface * /*netif*/)
{

    ELG(0, 0, 'dsbl', "org_lucidcake_driver_AJZaurusUSB::disable");

        // Disable our IOOutputQueue object. This will prevent the
        // outputPacket() method from being called
        
    fTransmitQueue->stop();

        // Flush all packets currently in the output queue

    fTransmitQueue->setCapacity(0);
    fTransmitQueue->flush();

    putToSleep();

    fNetifEnabled = false;
    fReady = false;

    return kIOReturnSuccess;
    
}/* end disable */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::setWakeOnMagicPacket
//
//		Inputs:		active - true(wake), false(don't)
//
//		Outputs:	Return code - kIOReturnSuccess
//
//		Desc:		Set for wake on magic packet
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::setWakeOnMagicPacket(bool active)
{
    IOUSBDevRequest	devreq;
    IOReturn		ior = kIOReturnSuccess;

    ELG(0, active, 'sWMP', "org_lucidcake_driver_AJZaurusUSB::setWakeOnMagicPacket");
	
    fWOL = active;
    
    if (fbmAttributes & kUSBAtrRemoteWakeup)
    {
    
            // Clear the feature if wake-on-lan is not set (SetConfiguration sets the feature 
            // automatically if the device supports remote wake up)
    
        if (!active)				
        {
            devreq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBStandard, kUSBDevice);
            devreq.bRequest = kUSBRqClearFeature;
            devreq.wValue = kUSBFeatureDeviceRemoteWakeup;
            devreq.wIndex = 0;
            devreq.wLength = 0;
            devreq.pData = 0;

            ior = fpDevice->DeviceRequest(&devreq);
            if (ior == kIOReturnSuccess)
            {
                ELG(0, ior, 'SCCs', "org_lucidcake_driver_AJZaurusUSB::initDevice - Clearing remote wake up feature successful");
            } else {
                ALERT(0, ior, 'SCCf', "org_lucidcake_driver_AJZaurusUSB::initDevice - Clearing remote wake up feature failed");
            }
        }
    } else {
        ELG(0, 0, 'SCRw', "org_lucidcake_driver_AJZaurusUSB::initDevice - Remote wake up not supported");
    }

    
    return kIOReturnSuccess;
    
}/* end setWakeOnMagicPacket */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::getPacketFilters
//
//		Inputs:		group - the filter group
//
//		Outputs:	Return code - kIOReturnSuccess and others
//				filters - the capability
//
//		Desc:		Set the filter capability for the driver
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::getPacketFilters(const OSSymbol *group, UInt32 *filters) const
{
    IOReturn	rtn = kIOReturnSuccess;
    
    ELG(group, filters, 'gPkF', "org_lucidcake_driver_AJZaurusUSB::getPacketFilters");

    if (group == gIOEthernetWakeOnLANFilterGroup)
    {
        if (fbmAttributes & kUSBAtrRemoteWakeup)
        {
            *filters = kIOEthernetWakeOnMagicPacket;
        } else {
            *filters = 0;
        }
    } else {
        if (group == gIONetworkFilterGroup)
        {
            *filters = kIOPacketFilterUnicast | kIOPacketFilterBroadcast | kIOPacketFilterMulticast | kIOPacketFilterMulticastAll | kIOPacketFilterPromiscuous;
        } else {
            rtn = super::getPacketFilters(group, filters);
        }
    }
    
    if (rtn != kIOReturnSuccess)
    {
        ALERT(0, rtn, 'gPk-', "org_lucidcake_driver_AJZaurusUSB::getPacketFilters - failed");
    }
    
    return rtn;
    
}/* end getPacketFilters */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::selectMedium
//
//		Inputs:
//
//		Outputs:
//
//		Desc:		Lets us know if someone is playing with ifconfig
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::selectMedium(const IONetworkMedium *medium)
{
    
    ELG(0, 0, 'SlMd', "org_lucidcake_driver_AJZaurusUSB::selectMedium");

    setSelectedMedium(medium);
    
    return kIOReturnSuccess;
        
}/* end selectMedium */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::getHardwareAddress
//
//		Inputs:		
//
//		Outputs:	Return code - kIOReturnSuccess or kIOReturnError
//				ea - the address
//
//		Desc:		Get the ethernet address from the hardware (actually the descriptor)
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::getHardwareAddress(IOEthernetAddress *ea)
{
    UInt32      i;

    ELG(0, 0, 'gHdA', "org_lucidcake_driver_AJZaurusUSB::getHardwareAddress");
     
    for (i=0; i<6; i++)
    {
        ea->bytes[i] = fEaddr[i];
    }

    return kIOReturnSuccess;
    
}/* end getHardwareAddress */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::newVendorString
//
//		Inputs:		
//
//		Outputs:	Return code - the vendor string
//
//		Desc:		Identifies the hardware vendor
//
/****************************************************************************************************/

const OSString* org_lucidcake_driver_AJZaurusUSB::newVendorString() const
{

    ELG(0, 0, 'nVSt', "org_lucidcake_driver_AJZaurusUSB::newVendorString");
    
    return OSString::withCString((const char *)defaultName);		// Maybe we should use the descriptors

}/* end newVendorString */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::newModelString
//
//		Inputs:		
//
//		Outputs:	Return code - the model string
//
//		Desc:		Identifies the hardware model
//
/****************************************************************************************************/

const OSString* org_lucidcake_driver_AJZaurusUSB::newModelString() const
{

    ELG(0, 0, 'nMSt', "org_lucidcake_driver_AJZaurusUSB::newModelString");
    
    return OSString::withCString("USB");		// Maybe we should use the descriptors
    
}/* end newModelString */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::newRevisionString
//
//		Inputs:		
//
//		Outputs:	Return code - the revision string
//
//		Desc:		Identifies the hardware revision
//
/****************************************************************************************************/

const OSString* org_lucidcake_driver_AJZaurusUSB::newRevisionString() const
{

    ELG(0, 0, 'nRSt', "org_lucidcake_driver_AJZaurusUSB::newRevisionString");
    
    return OSString::withCString("");
    
}/* end newRevisionString */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::setMulticastMode
//
//		Inputs:		active - true (set it), false (don't)
//
//		Outputs:	Return code - kIOReturnSuccess
//
//		Desc:		Sets multicast mode
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::setMulticastMode(bool active)
{

    ELG(0, active, 'stMM', "org_lucidcake_driver_AJZaurusUSB::setMulticastMode" );

    if (active)
    {
        fPacketFilter |= kPACKET_TYPE_ALL_MULTICAST;
    } else {
        fPacketFilter &= ~kPACKET_TYPE_ALL_MULTICAST;
    }
    
    USBSetPacketFilter();
    
    return kIOReturnSuccess;
    
}/* end setMulticastMode */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::setMulticastList
//
//		Inputs:		addrs - list of addresses
//				count - number in the list
//
//		Outputs:	Return code - kIOReturnSuccess or kIOReturnIOError
//
//		Desc:		Sets multicast list
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::setMulticastList(IOEthernetAddress *addrs, UInt32 count)
{
    bool	uStat;
    ELG(addrs, count, 'stML', "org_lucidcake_driver_AJZaurusUSB::setMulticastList" );
    
    if (count != 0)
    {
        uStat = USBSetMulticastFilter(addrs, count);
        if (!uStat)
        {
            return kIOReturnIOError;
        }
    }

    return kIOReturnSuccess;
    
}/* end setMulticastList */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::setPromiscuousMode
//
//		Inputs:		active - true (set it), false (don't)
//
//		Outputs:	Return code - kIOReturnSuccess
//
//		Desc:		Sets promiscuous mode
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::setPromiscuousMode(bool active)
{
    
    ALERT(0, active, 'stPM', "org_lucidcake_driver_AJZaurusUSB::setPromiscuousMode");

    if (active)
    {
        fPacketFilter |= kPACKET_TYPE_PROMISCUOUS;
    } else {
        fPacketFilter &= ~kPACKET_TYPE_PROMISCUOUS;
    }
    
    USBSetPacketFilter();
        
    return kIOReturnSuccess;
    
}/* end setPromiscuousMode */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::createOutputQueue
//
//		Inputs:		
//
//		Outputs:	Return code - the output queue
//
//		Desc:		Creates the output queue
//
/****************************************************************************************************/

IOOutputQueue* org_lucidcake_driver_AJZaurusUSB::createOutputQueue()
{

    ELG(0, 0, 'crOQ', "org_lucidcake_driver_AJZaurusUSB::createOutputQueue" );
    
    return IOBasicOutputQueue::withTarget(this, TRANSMIT_QUEUE_SIZE);
    
}/* end createOutputQueue */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::outputPacket
//
//		Inputs:		mbuf - the packet
//				param - optional parameter
//
//		Outputs:	Return code - kIOReturnOutputSuccess or kIOReturnOutputStall
//
//		Desc:		Packet transmission. The BSD mbuf needs to be formatted correctly
//				and transmitted
//
/****************************************************************************************************/

UInt32 org_lucidcake_driver_AJZaurusUSB::outputPacket(struct mbuf *pkt, void *param)
{
    UInt32	ret = kIOReturnOutputSuccess;
    
    ELG(pkt, 0, 'otPk', "org_lucidcake_driver_AJZaurusUSB::outputPacket" );

    if (!fLinkStatus)
    {
        ALERT(pkt, fLinkStatus, 'otL-', "org_lucidcake_driver_AJZaurusUSB::outputPacket - link is down" );
        if (fOutputErrsOK)
            fpNetStats->outputErrors++;
        freePacket(pkt);
    } else { 
        if (USBTransmitPacket(pkt) == false)
        {
            ret = kIOReturnOutputStall;
        }
    }

    return ret;
    
}/* end outputPacket */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::configureInterface
//
//		Inputs:		netif - the interface being configured
//
//		Outputs:	Return code - true (configured ok), false (not)
//
//		Desc:		Finish the network interface configuration
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::configureInterface(IONetworkInterface *netif)
{
    IONetworkData	*nd;

    ELG(IOThreadSelf(), netif, 'cfIt', "org_lucidcake_driver_AJZaurusUSB::configureInterface");

    if (super::configureInterface(netif) == false)
    {
        ALERT(0, 0, 'cfs-', "org_lucidcake_driver_AJZaurusUSB::configureInterface - super failed");
        return false;
    }
    
        // Get a pointer to the statistics structure in the interface

    nd = netif->getNetworkData(kIONetworkStatsKey);
    if (!nd || !(fpNetStats = (IONetworkStats *)nd->getBuffer()))
    {
        ALERT(0, 0, 'cfn-', "org_lucidcake_driver_AJZaurusUSB::configureInterface - Invalid network statistics");
        return false;
    }

        // Get the Ethernet statistics structure:

    nd = netif->getParameter(kIOEthernetStatsKey);
    if (!nd || !(fpEtherStats = (IOEthernetStats*)nd->getBuffer()))
    {
        ALERT(0, 0, 'cfe-', "org_lucidcake_driver_AJZaurusUSB::configureInterface - Invalid ethernet statistics\n" );
        return false;
    }

    return true;
    
}/* end configureInterface */

/****************************************************************************************************/
//
//		Method:		com_apple_driver_dts_USBCDCEthernet::wakeUp
//
//		Inputs:		
//
//		Outputs:	Return Code - true(we're awake), false(failed)
//
//		Desc:		Resumes the device it it was suspended and then gets all the data
//				structures sorted out and all the pipes ready.
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::wakeUp()
{
    IOReturn 	rtn = kIOReturnSuccess;

    ELG(0, 0, 'wkUp', "org_lucidcake_driver_AJZaurusUSB::wakeUp");
    
    fReady = false;
    
    if (fTimerSource)
    { 
        fTimerSource->cancelTimeout();
    }
    
    setLinkStatus(0, 0);				// Initialize the link state
    
    //if (fbmAttributes & kUSBAtrBusPowered)
    //{
    //    rtn = fpDevice->SuspendDevice(false);		// Resume the device
    //    if (rtn != kIOReturnSuccess)
    //    {
	//		ALERT(0, 0, 'wkS-', "org_lucidcake_driver_AJZaurusUSB::wakeUp - error waking device");
    //        return false;
    //    }
    //}
    //
    //IOSleep(50);
    //
	// AJ: The above is commented out because it's an endless source of
    // trouble (at least with the Zaurus, probably also with other
    // PDAs) - on Panther, it can even cause the USB subsystem to hang!
    // Not trying to suspend/wake up the Zaurus at all seems to be just
    // fine (it won't suspend anyway upon SuspendDevice(true)).
    
    if (!allocateResources()) 
    {
    	return false;
    }
		
        // Read the comm interrupt pipe for status:
		
    fCommCompletionInfo.target = this;
    fCommCompletionInfo.action = commReadComplete;
    fCommCompletionInfo.parameter = NULL;
    
    if (fCommPipe)
    {
        //rtn = fCommPipe->Read(fCommPipeMDP, &fCommCompletionInfo, NULL);
        rtn = kIOReturnSuccess;
    }
    if (rtn == kIOReturnSuccess)
    {
        	// Read the data-in bulk pipe:
			
        fReadCompletionInfo.target = this;
        fReadCompletionInfo.action = dataReadComplete;
        fReadCompletionInfo.parameter = NULL;
		
        rtn = fInPipe->Read(fPipeInMDP, &fReadCompletionInfo, NULL);
			
        if (rtn == kIOReturnSuccess)
        {
        	// Set up the data-out bulk pipe:
			
            //fWriteCompletionInfo.target	= this;
            //fWriteCompletionInfo.action	= dataWriteComplete;
            //fWriteCompletionInfo.parameter = NULL;				// for now, filled in with the mbuf address when sent

			for (int i=0; i<kOutBufPool; i++)
			{
				fPipeOutBuff[i].writeCompletionInfo.target = this;
				fPipeOutBuff[i].writeCompletionInfo.action = dataWriteComplete;
				fPipeOutBuff[i].writeCompletionInfo.parameter = NULL;				// for now, filled in with the mbuf address when sent
			}
	
                // Set up the management element request completion routine:

            fMERCompletionInfo.target = this;
            fMERCompletionInfo.action = merWriteComplete;
            fMERCompletionInfo.parameter = NULL;				// for now, filled in with parm block when allocated
            
                // Set up the statistics request completion routine:

            fStatsCompletionInfo.target = this;
            fStatsCompletionInfo.action = statsWriteComplete;
            fStatsCompletionInfo.parameter = NULL;				// for now, filled in with parm block when allocated
        }
    }

    if (rtn != kIOReturnSuccess)
    {
    
    	// We failed for some reason
	
        ALERT(0, 0, 'wkp-', "org_lucidcake_driver_AJZaurusUSB::wakeUp - Setting up the pipes failed" );
        releaseResources();
        return false;
    } else {
        if (!fMediumDict)
        {
            if (!createMediumTables())
            {
                ALERT(0, 0, 'wkc-', "org_lucidcake_driver_AJZaurusUSB::wakeUp - createMediumTables failed" );
                releaseResources();
                return false;
            }
        }

        fTimerSource->setTimeoutMS(WATCHDOG_TIMER_MS);
        fReady = true;
    }

    return true;
	
}/* end wakeUp */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::putToSleep
//
//		Inputs:		
//
//		Outputs:	Return Code - true(we're asleep), false(failed)
//
//		Desc:		Do clean up and suspend the device.
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::putToSleep()
{
    //IOReturn	ior;

    ELG(0, 0, 'pToS', "org_lucidcake_driver_AJZaurusUSB::putToSleep");
        
    fReady = false;

    if (fTimerSource)
    { 
        fTimerSource->cancelTimeout();
    }

    setLinkStatus(0, 0);
	
        // Release all resources
		
    releaseResources();
    
    //if (!fTerminate)
    //{
    //    if (fbmAttributes & kUSBAtrBusPowered)
    //    {
    //        ior = fpDevice->SuspendDevice(true);         // Suspend the device again (if supported and not unplugged)
    //        if (ior)
    //        {
    //            ALERT(0, ior, 'rPSD', "org_lucidcake_driver_AJZaurusUSB::releasePort - SuspendDevice error");
    //        }
    //    }
    //}
	// AJ: The above is commented out because it's an endless source of
    // trouble (at least with the Zaurus, probably also with other
    // PDAs).
    // Not trying to suspend/wake up the Zaurus at all seems to be just
    // fine (it won't suspend anyway upon SuspendDevice(true)).

    if ((fTerminate) && (!fReady))		// if it's the result of a terminate and no interfaces enabled we also need to close the device
    {
    	fpDevice->close(this);
        fpDevice = NULL;
    }
    
}/* end putToSleep */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::createMediumTables
//
//		Inputs:		
//
//		Outputs:	Return code - true (tables created), false (not created)
//
//		Desc:		Creates the medium tables
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::createMediumTables()
{
    IONetworkMedium	*medium;
    UInt64		maxSpeed;
    UInt32		i;

    ELG(0, 0, 'crMT', "org_lucidcake_driver_AJZaurusUSB::createMediumTables");

    maxSpeed = 100;
    fMediumDict = OSDictionary::withCapacity(sizeof(mediumTable) / sizeof(mediumTable[0]));
    if (fMediumDict == 0)
    {
        ALERT( 0, 0, 'crc-', "org_lucidcake_driver_AJZaurusUSB::createMediumTables - create dict. failed" );
        return false;
    }

    for (i = 0; i < sizeof(mediumTable) / sizeof(mediumTable[0]); i++ )
    {
        medium = IONetworkMedium::medium(mediumTable[i].type, mediumTable[i].speed);
        if (medium && (medium->getSpeed() <= maxSpeed))
        {
            IONetworkMedium::addMedium(fMediumDict, medium);
            medium->release();
        }
    }

    if (publishMediumDictionary(fMediumDict) != true)
    {
        ALERT( 0, 0, 'crp-', "org_lucidcake_driver_AJZaurusUSB::createMediumTables - publish dict. failed" );
        return false;
    }

    medium = IONetworkMedium::getMediumWithType(fMediumDict, kIOMediumEthernetAuto);
    setCurrentMedium(medium);

    return true;
    
}/* end createMediumTables */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::allocateResources
//
//		Inputs:		
//
//		Outputs:	return code - true (allocate was successful), false (it failed)
//
//		Desc:		Finishes up the rest of the configuration and gets all the endpoints open etc.
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::allocateResources()
{
    IOUSBFindEndpointRequest	epReq;		// endPoint request struct on stack
    UInt32			i;

	// AJ: init the sync counter
	// -->
	//fSyncCount = 0;
	// <--
	
    ELG(0, 0, 'Allo', "org_lucidcake_driver_AJZaurusUSB::allocateResources.");

	// --> debug: dump all the pipes
	//int endpointCount = fDataInterface->GetNumEndpoints();
	//ALERT(0, endpointCount, '----', "Enumerating pipes (alt. method) ...");
	//IOUSBPipe* enumPipe = NULL;
	//int counter = 0;
	//for (; counter < endpointCount; ++counter) {
	//	enumPipe = fDataInterface->GetPipeObj(counter);
	//	if (enumPipe == NULL) {
	//		ALERT(counter, 0, '----', "  Pipe is NULL");
	//	} else {
	//		ALERT(counter, 0, '----', "  Pipe found");
	//		ALERT(enumPipe->GetType(), enumPipe->GetDirection(), '----', "    type, direction");
	//		ALERT(enumPipe->GetMaxPacketSize(), enumPipe->GetInterval(), '----', "    maxPacketSize, interval");
	//	}
	//}
	/*do {
		epReq.type = kUSBAnyType;
		epReq.direction = kUSBAnyDirn;
		epReq.maxPacketSize	= 0;
		epReq.interval = 0;
		enumPipe = fDataInterface->FindNextPipe(enumPipe, &epReq);
		if (enumPipe != NULL) {
			ALERT(0, 0, '----', "  Pipe found");
			ALERT(epReq.type, epReq.direction, '----', "  type, direction");
			ALERT(epReq.maxPacketSize, epReq.interval, '----', "  maxPacketSize, interval");
		}
		++counter;
		if (counter > 10) {
			enumPipe = NULL;
		}
	} while (enumPipe != NULL);*/
	// <--
	
        // Open all the end points

    epReq.type = kUSBBulk;
    epReq.direction = kUSBIn;
    epReq.maxPacketSize	= 0;
    epReq.interval = 0;
    fInPipe = fDataInterface->FindNextPipe(0, &epReq);
    if (!fInPipe)
    {
        ALERT(0, 0, 'inP-', "org_lucidcake_driver_AJZaurusUSB::allocateResources - no bulk input pipe.");
        return false;
    }
    ELG(epReq.maxPacketSize << 16 |epReq.interval, fInPipe, 'inP+', "org_lucidcake_driver_AJZaurusUSB::allocateResources - bulk input pipe.");

    epReq.direction = kUSBOut;
    fOutPipe = fDataInterface->FindNextPipe(0, &epReq);
    if (!fOutPipe)
    {
        ALERT(0, 0, 'otP-', "org_lucidcake_driver_AJZaurusUSB::allocateResources - no bulk output pipe.");
        return false;
    }
    fOutPacketSize = epReq.maxPacketSize;
    ELG(epReq.maxPacketSize << 16 |epReq.interval, fOutPipe, 'otP+', "org_lucidcake_driver_AJZaurusUSB::allocateResources - bulk output pipe.");

        // Interrupt pipe - Comm Interface

    epReq.type = kUSBInterrupt;
    epReq.direction = kUSBIn;
    fCommPipe = fCommInterface->FindNextPipe(0, &epReq);
    if (!fCommPipe)
    {
        ELG(0, 0, 'cmP-', "org_lucidcake_driver_AJZaurusUSB::allocateResources - no interrupt in pipe.");
        fCommPipeMDP = NULL;
        fCommPipeBuffer = NULL;
        fLinkStatus = 1;					// Mark it active cause we'll never get told
//        return false;
    } else {
        ELG(epReq.maxPacketSize << 16 |epReq.interval, fCommPipe, 'cmP+', "org_lucidcake_driver_AJZaurusUSB::allocateResources - comm pipe.");

            // Allocate Memory Descriptor Pointer with memory for the Comm pipe:

        fCommPipeMDP = IOBufferMemoryDescriptor::withCapacity(COMM_BUFF_SIZE, kIODirectionIn);
        if (!fCommPipeMDP)
            return false;
		
        fCommPipeMDP->setLength(COMM_BUFF_SIZE);
        fCommPipeBuffer = (UInt8*)fCommPipeMDP->getBytesNoCopy();
        ELG(0, fCommPipeBuffer, 'cBuf', "org_lucidcake_driver_AJZaurusUSB::allocateResources - comm buffer");
    }

        // Allocate Memory Descriptor Pointer with memory for the data-in bulk pipe:

    fPipeInMDP = IOBufferMemoryDescriptor::withCapacity(fMax_Block_Size, kIODirectionIn);
    if (!fPipeInMDP)
        return false;
		
    fPipeInMDP->setLength(fMax_Block_Size);
    fPipeInBuffer = (UInt8*)fPipeInMDP->getBytesNoCopy();
    ELG(0, fPipeInBuffer, 'iBuf', "org_lucidcake_driver_AJZaurusUSB::allocateResources - input buffer");

        // Allocate Memory Descriptor Pointers with memory for the data-out bulk pipe pool

    for (i=0; i<kOutBufPool; i++)
    {
        fPipeOutBuff[i].pipeOutMDP = IOBufferMemoryDescriptor::withCapacity(fMax_Block_Size, kIODirectionOut);
        if (!fPipeOutBuff[i].pipeOutMDP)
        {
            ALERT(0, 0, 'obf-', "org_lucidcake_driver_AJZaurusUSB::allocateResources - Allocate output descriptor failed");
            return false;
        }
		
        fPipeOutBuff[i].pipeOutMDP->setLength(fMax_Block_Size);
        fPipeOutBuff[i].pipeOutBuffer = (UInt8*)fPipeOutBuff[i].pipeOutMDP->getBytesNoCopy();
        ELG(fPipeOutBuff[i].pipeOutMDP, fPipeOutBuff[i].pipeOutBuffer, 'oBuf', "org_lucidcake_driver_AJZaurusUSB::allocateResources - output buffer");
    }
		
    // AJ: debug info
	// --> debug	
    ELG(0, epReq.maxPacketSize, 'oBuf', "org_lucidcake_driver_AJZaurusUSB::allocateResources - max packet size");
	// <--
		
    return true;
	
}/* end allocateResources */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::releaseResources
//
//		Inputs:		
//
//		Outputs:	
//
//		Desc:		Frees up the resources allocated in allocateResources
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::releaseResources()
{
    UInt32	i;
    
    ELG(0, 0, 'rlRs', "org_lucidcake_driver_AJZaurusUSB::releaseResources");

    for (i=0; i<kOutBufPool; i++)
    {
        if (fPipeOutBuff[i].pipeOutMDP)	
        { 
            fPipeOutBuff[i].pipeOutMDP->release();	
            fPipeOutBuff[i].pipeOutMDP = NULL;
        }
    }
	
    if (fPipeInMDP  )	
    { 
        fPipeInMDP->release();	
        fPipeInMDP = 0; 
    }
	
    if (fCommPipeMDP)	
    { 
        fCommPipeMDP->release();	
        fCommPipeMDP = 0; 
    }
	
}/* end releaseResources */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket
//
//		Inputs:		packet - the packet
//
//		Outputs:	Return code - true (transmit started), false (it didn't)
//
//		Desc:		Set up and then transmit the packet
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket(struct mbuf *packet)
{
    UInt32		numbufs = 0;		// number of mbufs for this packet
    struct mbuf		*m;				// current mbuf
    UInt32		total_pkt_length = 0;
	// AJ: new local variables
	// -->
	UInt32		new_pkt_length;
	UInt32		checksum_length;
	UInt32		fcs;
	UInt32		pad;
	// <--
    UInt32		rTotal = 0;
    IOReturn		ior = kIOReturnSuccess;
    UInt32		poolIndx;
    bool		gotBuffer = false;
    UInt16		tryCount = 0;
	
    ELG (0, packet, 'txPk', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket");

	//if (OSDecrementAtomic(&fSyncCount) < 0) {
	//	ELG(0, fSyncCount, 'txPr', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - packet rejected");
	//	// AJ: TODO add a ring buffer or something to help performance
	//	OSIncrementAtomic(&fSyncCount);
	//	return false;
	//}

	// Count the number of mbufs in this packet
        
    m = packet;
    while (m)
    {
        total_pkt_length += m->m_len;
        numbufs++;
	    m = m->m_next;
    }
    
    // AJ: Calculate size including CRC and padding
	// -->
	if (fChecksum) {
		checksum_length = 4;
	} else {
		checksum_length = 0;
	}
    if (fPadded) {
		// we need to pad so that after appending the CRC we have a multiple of packetsize less one
		new_pkt_length = fOutPacketSize * ( ((total_pkt_length + checksum_length) / fOutPacketSize) + 1) - 1;
	} else {
		if (fChecksum) {
			// require a minimum of one full packet
			new_pkt_length = MAX(fOutPacketSize, total_pkt_length + checksum_length);
		} else {
			new_pkt_length = total_pkt_length;
		}
	}
    // <--
	
    ELG(new_pkt_length, numbufs, 'txTN', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - Total packet length and Number of mbufs");
    
    if (new_pkt_length+1 > fMax_Block_Size)
    {
        ALERT(0, 0, 'txBp', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - Bad packet size");	// Note for now and revisit later
        if (fOutputErrsOK)
            fpNetStats->outputErrors++;
		//OSIncrementAtomic(&fSyncCount);
		return false;
	}
    
            // Find an ouput buffer in the pool
    
	IOSimpleLockLock(fLock);
    while (!gotBuffer)
    {
        for (poolIndx=0; poolIndx<kOutBufPool; poolIndx++)
        {
            if (fPipeOutBuff[poolIndx].m == NULL)
            {
                gotBuffer = true;
                break;
            }
        }
        if (gotBuffer)
        {
            break;
        } else {
			// --> crash test
			//gotBuffer = true;
			//poolIndx = 0;
			//break;
			// <--
            tryCount++;
            if (tryCount > kOutBuffThreshold)
            {
                ELG(0, 0, 'txBT', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - Exceeded output buffer wait threshold");
                if (fOutputErrsOK)
                    fpNetStats->outputErrors++;
				IOSimpleLockUnlock(fLock);
				fOutputStalled = true;
                return false;
            } else {
                ELG(0, tryCount, 'txBT', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - Waiting for output buffer");
                IOSleep(1);
            }
        }
    }
	fPipeOutBuff[poolIndx].m = packet;
	++fDataCount;
	if (fDataCount < 4) {
		//ALERT(IOThreadSelf(), fDataCount, '----', "Data count (low)");
	} else {
		//ALERT(IOThreadSelf(), fDataCount, '----', "Data count (high)");
	}
	fOutputStalled = false;
	IOSimpleLockUnlock(fLock);

    // Start filling in the send buffer
	// AJ: modified to do CRC calculation and padding
	// -->
	fcs = CRC32_INITFCS;
    m = packet;							// start with the first mbuf of the packet
    rTotal = 0;							// running total				
    do
    {  
        if (m->m_len == 0)					// Ignore zero length mbufs
            continue;
        fcs = fcs_memcpy32(&(fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal]), mtod(m, unsigned char*), m->m_len, fcs);
        //bcopy(mtod(m, unsigned char *), &fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal], m->m_len);
        rTotal += m->m_len;
        
    } while ((m = m->m_next) != 0);
    if ((pad = new_pkt_length - rTotal - checksum_length) > 0) {
        // pad to required length less four (CRC), copy fcs and append pad byte if required
        fcs = fcs_pad32(&(fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal]), pad, fcs);
		rTotal += pad;
    }
	fcs = ~fcs;
	if (fChecksum)
	{
		fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal++] = fcs&0xff;
		fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal++] = (fcs>>8)&0xff;
		fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal++] = (fcs>>16)&0xff;
		fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal++] = (fcs>>24)&0xff;
	}

    // append a byte if required, we overallocated by one to allow for this
	if (fPadded)
	{
		if (!(rTotal % fOutPacketSize)) {
			fPipeOutBuff[poolIndx].pipeOutBuffer[rTotal++] = 0;
		}
	}
    ELG(rTotal, fcs, 'txLe', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - length+checksum");
	// <--
    
    LogData(kUSBOut, rTotal, fPipeOutBuff[poolIndx].pipeOutBuffer);

	//fPipeOutBuff[poolIndx].m = packet;
    fPipeOutBuff[poolIndx].writeCompletionInfo.parameter = (void *)poolIndx;
	// FIXME: completion info should be part of the pool!
    fPipeOutBuff[poolIndx].pipeOutMDP->setLength(rTotal);
    ior = fOutPipe->Write(fPipeOutBuff[poolIndx].pipeOutMDP, &(fPipeOutBuff[poolIndx].writeCompletionInfo));
    if (ior != kIOReturnSuccess)
    {
		//OSIncrementAtomic(&fSyncCount);
        ALERT((UInt32) (ior == kIOUSBPipeStalled), ior, 'tx1-', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - Write failed");
        if (ior == kIOUSBPipeStalled)
        {
			//OSDecrementAtomic(&fSyncCount);
            fOutPipe->Reset();
            ior = fOutPipe->Write(fPipeOutBuff[poolIndx].pipeOutMDP, &(fPipeOutBuff[poolIndx].writeCompletionInfo));
            if (ior != kIOReturnSuccess)
            {
				//OSIncrementAtomic(&fSyncCount);
                ALERT(0, ior, 'tx2-', "org_lucidcake_driver_AJZaurusUSB::USBTransmitPacket - Write really failed");
                if (fOutputErrsOK)
                    fpNetStats->outputErrors++;
				IOSimpleLockLock(fLock);
				--fDataCount;
				fPipeOutBuff[poolIndx].m = NULL;
				IOSimpleLockUnlock(fLock);
                return false;
            }
		} else {
			IOSimpleLockLock(fLock);
			--fDataCount;
			IOSimpleLockUnlock(fLock);
			fPipeOutBuff[poolIndx].m = NULL;
            return false;
		}
	}

    freePacket(packet);
	if (fOutputPktsOK)
        fpNetStats->outputPackets++;
    
    return true;

}/* end USBTransmitPacket */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter
//
//		Inputs:		addrs - the list of addresses
//				count - How many
//
//		Outputs:	
//
//		Desc:		Set up and send SetMulticastFilter Management Element Request(MER).
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter(IOEthernetAddress *addrs, UInt32 count)
{
    IOReturn		rc;
    IOUSBDevRequest	*MER;
    UInt8		*eaddrs;
    UInt32		eaddLen;
    UInt32		i,j,rnum;
	
    ELG(fMcFilters, count, 'USMF', "org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter");

    if (count > (UInt32)(fMcFilters & kFiltersSupportedMask))
    {
        ELG(0, 0, 'USf-', "org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter - No multicast filters supported");
        return false;
    }

    MER = (IOUSBDevRequest*)IOMalloc(sizeof(IOUSBDevRequest));
    if (!MER)
    {
        ALERT(0, 0, 'USM-', "org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter - allocate MER failed");
        return false;
    }
    bzero(MER, sizeof(IOUSBDevRequest));
	
    eaddLen = count * kIOEthernetAddressSize;
    eaddrs = (UInt8 *)IOMalloc(eaddLen);
    if (!eaddrs)
    {
        ALERT(0, 0, 'USA-', "org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter - allocate address buffer failed");
        return false;
    }
    bzero(eaddrs, eaddLen); 
	
        // Build the filter address buffer
         
    rnum = 0;
    for (i=0; i<count; i++)
    {
        if (rnum > eaddLen)				// Just in case
        {
            break;
        }
        for (j=0; j<kIOEthernetAddressSize; j++)
        {
            eaddrs[rnum++] = addrs->bytes[j];
        }
    }
    
        // Now build the Management Element Request
		
    MER->bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
    MER->bRequest = kSet_Ethernet_Multicast_Filter;
    MER->wValue = count;
    MER->wIndex = fCommInterfaceNumber;
    MER->wLength = eaddLen;
    MER->pData = eaddrs;
	
    fMERCompletionInfo.parameter = MER;
	
    rc = fpDevice->DeviceRequest(MER, &fMERCompletionInfo);
    if (rc != kIOReturnSuccess)
    {
        ALERT(MER->bRequest, rc, 'USE-', "org_lucidcake_driver_AJZaurusUSB::USBSetMulticastFilter - Error issueing DeviceRequest");
        IOFree(MER->pData, eaddLen);
        IOFree(MER, sizeof(IOUSBDevRequest));
        return false;
    }
    
    return true;

}/* end USBSetMulticastFilter */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::USBSetPacketFilter
//
//		Inputs:		
//
//		Outputs:	
//
//		Desc:		Set up and send SetEthernetPackettFilters Management Element Request(MER).
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::USBSetPacketFilter()
{
    IOReturn		rc;
    IOUSBDevRequest	*MER;
	
    ELG(0, fPacketFilter, 'USPF', "org_lucidcake_driver_AJZaurusUSB::USBSetPacketFilter");
	
    MER = (IOUSBDevRequest*)IOMalloc(sizeof(IOUSBDevRequest));
    if (!MER)
    {
        ALERT(0, 0, 'USP-', "org_lucidcake_driver_AJZaurusUSB::USBSetPacketFilter - allocate MER failed");
        return false;
    }
    bzero(MER, sizeof(IOUSBDevRequest));
    
        // Now build the Management Element Request
		
    MER->bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
    MER->bRequest = kSet_Ethernet_Packet_Filter;
    MER->wValue = fPacketFilter;
    MER->wIndex = fCommInterfaceNumber;
    MER->wLength = 0;
    MER->pData = NULL;
	
    fMERCompletionInfo.parameter = MER;
	
    rc = fpDevice->DeviceRequest(MER, &fMERCompletionInfo);
    if (rc != kIOReturnSuccess)
    {
        ALERT(MER->bRequest, rc, 'USE-', "org_lucidcake_driver_AJZaurusUSB::USBSetPacketFilter - DeviceRequest error");
        if (rc == kIOUSBPipeStalled)
        {

            // Clear the stall and try it once more
        
            fpDevice->GetPipeZero()->ClearPipeStall(false);
            rc = fpDevice->DeviceRequest(MER, &fMERCompletionInfo);
            if (rc != kIOReturnSuccess)
            {
                ALERT(MER->bRequest, rc, 'USE-', "org_lucidcake_driver_AJZaurusUSB::USBSetPacketFilter - DeviceRequest, error a second time");
                IOFree(MER, sizeof(IOUSBDevRequest));
                return false;
            }
        }
    }
    
    return true;
    
}/* end USBSetPacketFilter */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::clearPipeStall
//
//		Inputs:		thePipe - the pipe
//
//		Outputs:	
//
//		Desc:		Clear a stall on the specified pipe. All outstanding I/O
//				is returned as aborted.
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::clearPipeStall(IOUSBPipe *thePipe)
{
    //UInt8	pipeStatus;
    IOReturn 	rtn = kIOReturnSuccess;
    
    ALERT(0, thePipe, 'clSt', "org_lucidcake_driver_AJZaurusUSB::clearPipeStall");
    
    /*pipeStatus = thePipe->GetStatus();
    if (pipeStatus == kPipeStalled)
    {
        rtn = thePipe->ClearPipeStall(true);
        if (rtn == kIOReturnSuccess)
        {
            ELG(0, 0, 'clSS', "org_lucidcake_driver_AJZaurusUSB::clearPipeStall - Successful");
        } else {
            ALERT(0, rtn, 'clSF', "org_lucidcake_driver_AJZaurusUSB::clearPipeStall - Failed");
        }
    } else {
        ELG(0, pipeStatus, 'clSP', "org_lucidcake_driver_AJZaurusUSB::clearPipeStall - Pipe not stalled");
    }*/
	
	// rewritten to prevent nasty issues when doing synchronous stuff here
	rtn = thePipe->ClearPipeStall(false);
    
    return rtn;

}/* end clearPipeStall */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::receivePacket
//
//		Inputs:		packet - the packet
//				size - Number of bytes in the packet
//
//		Outputs:	
//
//		Desc:		Build the mbufs and then send to the network stack.
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::receivePacket(UInt8 *packet, UInt32 size)
{
    struct mbuf		*m;
    UInt32		submit;
	UInt32		fcs;
    
    ELG(0, size, 'rcPk', "org_lucidcake_driver_AJZaurusUSB::receivePacket");
    
    if (size > fMax_Block_Size)
    {
        ALERT(0, 0, 'rcP-', "org_lucidcake_driver_AJZaurusUSB::receivePacket - Packet size error, packet dropped");
        if (fInputErrsOK)
            fpNetStats->inputErrors++;
        return;
    }

	// check CRC
	
	if (fChecksum) {
		if ((size % fOutPacketSize) == 1) {
	
			// check fcs across length minus one bytes
			if ((fcs = fcs_compute32(packet, size - 1, CRC32_INITFCS)) == CRC32_GOODFCS) {
				// success, trim extra byte and fall through
				--size;
			}
			// failed, check additional byte
			else if ((fcs = fcs_compute32(packet+size - 1, 1, fcs)) != CRC32_GOODFCS) {
				// failed
				ALERT(size, fcs, 'cs1-', "org_lucidcake_driver_AJZaurusUSB::receivePacket - CRC fail on exta byte, packet dropped");
				if (fInputErrsOK)
					fpNetStats->inputErrors++;
				return;
			}
			// success fall through, possibly with corrected length
		}
		// normal check across full frame
		else if ((fcs = fcs_compute32(packet, size, CRC32_INITFCS)) != CRC32_GOODFCS) {
			ALERT(size, fcs, 'cs2-', "org_lucidcake_driver_AJZaurusUSB::receivePacket - CRC failed, packet dropped");
			if (fInputErrsOK)
				fpNetStats->inputErrors++;
			return;
		}

		// trim fcs
		size -= 4;
	}

	// push the packet up the stack
    m = allocatePacket(size);
    if (m)
    {
        bcopy(packet, mtod(m, unsigned char *), size);
        submit = fNetworkInterface->inputPacket(m, size);
        ELG(0, submit, 'rcSb', "org_lucidcake_driver_AJZaurusUSB::receivePacket - Packets submitted");
        if (fInputPktsOK)
            fpNetStats->inputPackets++;
    } else {
        ALERT(0, 0, 'rcB-', "org_lucidcake_driver_AJZaurusUSB::receivePacket - Buffer allocation failed, packet dropped");
        if (fInputErrsOK)
            fpNetStats->inputErrors++;
    }

}/* end receivePacket */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::timerFired
//
//		Inputs:		
//
//		Outputs:	
//
//		Desc:		Static member function called when a timer event fires.
//
/****************************************************************************************************/
void org_lucidcake_driver_AJZaurusUSB::timerFired(OSObject *owner, IOTimerEventSource *sender)
{

//    ELG(0, 0, 'tmFd', "com_apple_driver_dts_USBCDCEthernet::timerFired");
    
    if (owner)
    {
	org_lucidcake_driver_AJZaurusUSB* target = OSDynamicCast(org_lucidcake_driver_AJZaurusUSB, owner);
	
	if (target)
	{
	    target->timeoutOccurred(sender);
	}
    }
    
}/* end timerFired */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::timeoutOccurred
//
//		Inputs:		
//
//		Outputs:	
//
//		Desc:		Timeout handler, used for stats gathering.
//
/****************************************************************************************************/

void org_lucidcake_driver_AJZaurusUSB::timeoutOccurred(IOTimerEventSource * /*timer*/)
{
    UInt32		*enetStats;
    UInt16		currStat;
    IOReturn		rc;
    IOUSBDevRequest	*STREQ;
    bool		statOk = false;
    
    ELG(0, 0, 'tmOd', "org_lucidcake_driver_AJZaurusUSB::timeoutOccurred");

	// AJ: revive a stalled queue (according to Apple's documentation,
	// this call doesn't do any harm, even if the queue wasn't stalled).
	// -->
	if (fReady) {
		fTransmitQueue->service(IOBasicOutputQueue::kServiceAsync);
	}
	// <--

    enetStats = (UInt32 *)&fEthernetStatistics;
    if (*enetStats == 0)
    {
        ELG(0, 0, 'tmN-', "org_lucidcake_driver_AJZaurusUSB::timeoutOccurred - No Ethernet statistics defined");
		fTimerSource->setTimeoutMS(WATCHDOG_TIMER_MS);	// AJ: set up the watchdog again
        return;
    }
    
    if (fReady == false)
    {
        ELG(0, 0, 'tmS-', "org_lucidcake_driver_AJZaurusUSB::timeoutOccurred - Spurious");    
    } else {
    
            // Only do it if it's not already in progress
    
        if (!fStatInProgress)
        {

                // Check if the stat we're currently interested in is supported
            
            currStat = stats[fCurrStat++];
            if (fCurrStat >= numStats)
            {
                fCurrStat = 0;
            }
            switch(currStat)
            {
                case kXMIT_OK_REQ:
                    if (fEthernetStatistics[0] & kXMIT_OK)
                    {
                        statOk = true;
                    }
                    break;
                case kRCV_OK_REQ:
                    if (fEthernetStatistics[0] & kRCV_OK)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_ERROR_REQ:
                    if (fEthernetStatistics[0] & kXMIT_ERROR_REQ)
                    {
                        statOk = true;
                    }
                    break;
                case kRCV_ERROR_REQ:
                    if (fEthernetStatistics[0] & kRCV_ERROR_REQ)
                    {
                        statOk = true;
                    }
                    break;
                case kRCV_CRC_ERROR_REQ:
                    if (fEthernetStatistics[2] & kRCV_CRC_ERROR)
                    {
                        statOk = true;
                    }
                    break;
                case kRCV_ERROR_ALIGNMENT_REQ:
                    if (fEthernetStatistics[2] & kRCV_ERROR_ALIGNMENT)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_ONE_COLLISION_REQ:
                    if (fEthernetStatistics[2] & kXMIT_ONE_COLLISION)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_MORE_COLLISIONS_REQ:
                    if (fEthernetStatistics[2] & kXMIT_MORE_COLLISIONS)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_DEFERRED_REQ:
                    if (fEthernetStatistics[2] & kXMIT_DEFERRED)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_MAX_COLLISION_REQ:
                    if (fEthernetStatistics[2] & kXMIT_MAX_COLLISION)
                    {
                        statOk = true;
                    }
                    break;
                case kRCV_OVERRUN_REQ:
                    if (fEthernetStatistics[3] & kRCV_OVERRUN)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_TIMES_CARRIER_LOST_REQ:
                    if (fEthernetStatistics[3] & kXMIT_TIMES_CARRIER_LOST)
                    {
                        statOk = true;
                    }
                    break;
                case kXMIT_LATE_COLLISIONS_REQ:
                    if (fEthernetStatistics[3] & kXMIT_LATE_COLLISIONS)
                    {
                        statOk = true;
                    }
                    break;
                default:
                    break;
            }
        }

        if (statOk)
        {
            STREQ = (IOUSBDevRequest*)IOMalloc(sizeof(IOUSBDevRequest));
            if (!STREQ)
            {
                ALERT(0, 0, 'tma-', "org_lucidcake_driver_AJZaurusUSB::timeoutOccurred - allocate STREQ failed");
            } else {
                bzero(STREQ, sizeof(IOUSBDevRequest));
        
                    // Now build the Statistics Request
		
                STREQ->bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
                STREQ->bRequest = kGet_Ethernet_Statistics;
                STREQ->wValue = currStat;
                STREQ->wIndex = fCommInterfaceNumber;
                STREQ->wLength = 4;
                STREQ->pData = &fStatValue;
	
                fStatsCompletionInfo.parameter = STREQ;
	
                rc = fpDevice->DeviceRequest(STREQ, &fStatsCompletionInfo);
                if (rc != kIOReturnSuccess)
                {
                    ALERT(STREQ->bRequest, rc, 'tmE-', "org_lucidcake_driver_AJZaurusUSB::timeoutOccurred - Error issueing DeviceRequest");
                    IOFree(STREQ, sizeof(IOUSBDevRequest));
                } else {
                    fStatInProgress = true;
                }
            }
        }
    }

        // Restart the watchdog timer
        
    fTimerSource->setTimeoutMS(WATCHDOG_TIMER_MS);

}/* end timeoutOccurred */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::message
//
//		Inputs:		type - message type, provider - my provider, argument - additional parameters
//
//		Outputs:	return Code - kIOReturnSuccess
//
//		Desc:		Handles IOKit messages. 
//
/****************************************************************************************************/

IOReturn org_lucidcake_driver_AJZaurusUSB::message(UInt32 type, IOService *provider, void *argument)
{
    IOReturn	ior;
	
    ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message");
	
    switch (type)
    {
        case kIOMessageServiceIsTerminated:
            ELG(fReady, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOMessageServiceIsTerminated");
			
            if (fReady)
            {
                if (!fTerminate)		// Check if we're already being terminated
                { 
            ALERT(0, 0, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - hardcoded message!");
		    // NOTE! This call below depends on the hard coded path of this KEXT. Make sure
		    // that if the KEXT moves, this path is changed!
		    KUNCUserNotificationDisplayNotice(
			0,		// Timeout in seconds
			0,		// Flags (for later usage)
			"",		// iconPath (not supported yet)
			"",		// soundPath (not supported yet)
			"",		// localizationPath (AJ: removed hard coded path, rather not support it)
			"Unplug Header",		// the header
			"Unplug Notice",		// the notice - look in Localizable.strings
			"OK"); 
                }
            } else {
                if (fCommInterface)	
                {
					if (!fSingleInterface) {
						fCommInterface->close(this);
					}
                    fCommInterface->release();
                    fCommInterface = NULL;	
                }
	
                if (fDataInterface)	
                { 
                    fDataInterface->close(this);	
                    fDataInterface->release();
                    fDataInterface = NULL;	
                }
                
            	fpDevice->close(this); 	// need to close so we can get the free and stop calls, only if no sessions active (see putToSleep)
                fpDevice = NULL;
            }
			
            fTerminate = true;		// we're being terminated (unplugged)
            return kIOReturnSuccess;			
        case kIOMessageServiceIsSuspended: 	
            ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOMessageServiceIsSuspended");
            break;			
        case kIOMessageServiceIsResumed: 	
            ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOMessageServiceIsResumed");
            break;			
        case kIOMessageServiceIsRequestingClose: 
            ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOMessageServiceIsRequestingClose"); 
            break;
        case kIOMessageServiceWasClosed: 	
            ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOMessageServiceWasClosed"); 
            break;
        case kIOMessageServiceBusyStateChange: 	
            ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOMessageServiceBusyStateChange"); 
            break;
        case kIOUSBMessagePortHasBeenResumed: 	
            ALERT(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOUSBMessagePortHasBeenResumed");
            
            // If the reads are dead try and resurrect them
            
			if (fCommDead)
			{
				ior = fCommPipe->Read(fCommPipeMDP, &fCommCompletionInfo, NULL);
				if (ior != kIOReturnSuccess)
				{
					ALERT(0, ior, 'msC-', "org_lucidcake_driver_AJZaurusUSB::message - Failed to queue Comm pipe read");
				} else {
					fCommDead = false;
				}
			}
            
			if (fDataDead)
			{
				ior = fInPipe->Read(fPipeInMDP, &fReadCompletionInfo, NULL);
				if (ior != kIOReturnSuccess)
				{
					ALERT(0, ior, 'msD-', "org_lucidcake_driver_AJZaurusUSB::message - Failed to queue Data pipe read");
				} else {
					fDataDead = false;
				}
			}

            break;
        case kIOUSBMessageHubResumePort:
            ELG(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - kIOUSBMessageHubResumePort");
            break;
        default:
            ALERT(0, type, 'mess', "org_lucidcake_driver_AJZaurusUSB::message - unknown message"); 
            break;
    }
    
    return kIOReturnUnsupported;
}/* end message */

/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::createWorkLoop
//
//		Inputs:
//
//		Outputs:	return Code - true if the work loop could be created, false otherwise.
//
//		Desc:		Creates a work loop for the driver. 
//
/****************************************************************************************************/

bool org_lucidcake_driver_AJZaurusUSB::createWorkLoop()
{
	fWorkLoop = IOWorkLoop::workLoop();
    ELG(fWorkLoop->getThread(), fWorkLoop, 'wrkl', "org_lucidcake_driver_AJZaurusUSB::createWorkLoop");
	
	return (fWorkLoop != 0);
}/* end createWorkLoop */


/****************************************************************************************************/
//
//		Method:		org_lucidcake_driver_AJZaurusUSB::getWorkLoop
//
//		Inputs:
//
//		Outputs:	return Code - the work loop.
//
//		Desc:		Returns the driver's work loop. 
//
/****************************************************************************************************/

IOWorkLoop* org_lucidcake_driver_AJZaurusUSB::getWorkLoop() const
{
    ELG(0, fWorkLoop, 'getl', "org_lucidcake_driver_AJZaurusUSB::getWorkLoop");
	return fWorkLoop;
}/* end getWorkLoop */
